# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Explorer Main Widget.
"""

# Standard library imports
import logging
import os.path as osp

# Third-party imports
from qtpy.QtCore import Qt, Signal, Slot
from qtpy.QtWidgets import (
    QDialog,
    QDialogButtonBox,
    QHBoxLayout,
    QLabel,
    QStackedWidget,
    QTextEdit,
    QVBoxLayout,
    QWidget,
)

# Local imports
from spyder.api.asyncdispatcher import AsyncDispatcher
from spyder.api.translations import _
from spyder.api.widgets.dialogs import SpyderDialogButtonBox
from spyder.api.widgets.main_widget import PluginMainWidget
from spyder.config.main import NAME_FILTERS
from spyder.plugins.explorer.widgets.remote_explorer import RemoteExplorer
from spyder.plugins.explorer.widgets.explorer import (
    DirViewActions,
    ExplorerTreeWidget,
)
from spyder.utils.icon_manager import ima
from spyder.utils.misc import getcwd_or_home

logger = logging.getLogger(__name__)


# ---- Constants
class ExplorerWidgetOptionsMenuSections:
    Files = 'files_section'
    Header = 'header_section'
    Common = 'common_section'


class ExplorerWidgetMainToolbarSections:
    Main = 'main_section'


class ExplorerWidgetActions:
    # Toggles
    ToggleFilter = 'toggle_filter_files_action'

    # Triggers
    Next = 'next_action'
    Parent = 'parent_action'
    Previous = 'previous_action'
    Refresh = 'refresh_action'
    EditNameFilters = 'edit_name_filters_action'


# ---- Main widget
class ExplorerWidget(PluginMainWidget):
    """Explorer widget"""

    # --- Signals
    # ------------------------------------------------------------------------
    sig_dir_opened = Signal(str, str)
    """
    This signal is emitted to indicate a folder has been opened.

    Parameters
    ----------
    directory: str
        The path to the directory opened.
    server_id: str
        The server identification from where the directory path is reachable.
    """

    sig_module_created = Signal(str)
    """
    This signal is emitted when a new python module is created.

    Parameters
    ----------
    module: str
        Path to the new module created.
    """

    sig_file_created = Signal(str)
    """
    This signal is emitted to request creating a new file with Spyder.

    Parameters
    ----------
    path: str
        File path to create.
    """

    sig_open_file_requested = Signal(str)
    """
    This signal is emitted to request opening a new file with Spyder.

    Parameters
    ----------
    path: str
        File path to run.
    """

    sig_removed = Signal(str)
    """
    This signal is emitted when a file is removed.

    Parameters
    ----------
    path: str
        File path removed.
    """

    sig_renamed = Signal(str, str)
    """
    This signal is emitted when a file is renamed.

    Parameters
    ----------
    old_path: str
        Old path for renamed file.
    new_path: str
        New path for renamed file.
    """

    sig_tree_removed = Signal(str)
    """
    This signal is emitted when a folder is removed.

    Parameters
    ----------
    path: str
        Folder to remove.
    """

    sig_tree_renamed = Signal(str, str)
    """
    This signal is emitted when a folder is renamed.

    Parameters
    ----------
    old_path: str
        Old path for renamed folder.
    new_path: str
        New path for renamed folder.
    """

    sig_run_requested = Signal(str)
    """
    This signal is emitted to request running a file.

    Parameters
    ----------
    path: str
        File path to run.
    """

    sig_open_interpreter_requested = Signal(str)
    """
    This signal is emitted to request opening an interpreter with the given
    path as working directory.

    Parameters
    ----------
    path: str
        Path to use as working directory of interpreter.
    """

    ENABLE_SPINNER = True

    def __init__(self, name, plugin, parent=None):
        """
        Initialize the widget.

        Parameters
        ----------
        name: str
            Name of the container.
        plugin: SpyderDockablePlugin
            Plugin of the container
        parent: QWidget
            Parent of this widget
        """
        super().__init__(name, plugin=plugin, parent=parent)

        # Widgets
        self.stackwidget = QStackedWidget(parent=self)
        self.treewidget = ExplorerTreeWidget(parent=self)
        self.remote_treewidget = RemoteExplorer(parent=self)
        self.stackwidget.addWidget(self.remote_treewidget)
        self.stackwidget.addWidget(self.treewidget)

        # Layouts
        layout = QHBoxLayout()
        layout.addWidget(self.stackwidget)
        self.setLayout(layout)

        # Signals
        self.treewidget.sig_dir_opened.connect(self.sig_dir_opened)
        self.remote_treewidget.sig_dir_opened.connect(self.sig_dir_opened)
        self.remote_treewidget.sig_start_spinner_requested.connect(
            self.start_spinner
        )
        self.remote_treewidget.sig_stop_spinner_requested.connect(
            self.stop_spinner
        )
        self.treewidget.sig_file_created.connect(self.sig_file_created)
        self.treewidget.sig_open_file_requested.connect(
            self.sig_open_file_requested)
        self.treewidget.sig_module_created.connect(self.sig_module_created)
        self.treewidget.sig_open_interpreter_requested.connect(
            self.sig_open_interpreter_requested)
        self.treewidget.sig_renamed.connect(self.sig_renamed)
        self.treewidget.sig_removed.connect(self.sig_removed)
        self.treewidget.sig_run_requested.connect(self.sig_run_requested)
        self.treewidget.sig_tree_removed.connect(self.sig_tree_removed)
        self.treewidget.sig_tree_renamed.connect(self.sig_tree_renamed)
        self.treewidget.sig_redirect_stdio_requested.connect(
            self.sig_redirect_stdio_requested)

    # ---- PluginMainWidget API
    # ------------------------------------------------------------------------
    def get_focus_widget(self):
        """Define the widget to focus."""
        return self.treewidget

    def get_title(self):
        """Return the title of the plugin tab."""
        return _("Files")

    def setup(self):
        """Perform the setup of the plugin's menu and actions."""
        # Common actions
        self.previous_action = self.create_action(
            ExplorerWidgetActions.Previous,
            text=_("Previous"),
            icon=self.create_icon('previous'),
            triggered=self.go_to_previous_directory,
        )
        self.next_action = self.create_action(
            ExplorerWidgetActions.Next,
            text=_("Next"),
            icon=self.create_icon('next'),
            triggered=self.go_to_next_directory,
        )
        self.parent_action = self.create_action(
            ExplorerWidgetActions.Parent,
            text=_("Parent"),
            icon=self.create_icon('up'),
            triggered=self.go_to_parent_directory
        )
        self.refresh_action = self.create_action(
            ExplorerWidgetActions.Refresh,
            text=_("Refresh"),
            icon=self.create_icon('refresh'),
            triggered=lambda: self.refresh(force_current=True)
        )
        filters_action = self.create_action(
            ExplorerWidgetActions.EditNameFilters,
            text=_("Edit filter settings..."),
            icon=self.create_icon('filter'),
            triggered=self.edit_filter,
        )

        # Common toolbuttons
        self.filter_button = self.create_action(
            ExplorerWidgetActions.ToggleFilter,
            text="",
            icon=ima.icon('filter'),
            toggled=self.change_filter_state
        )
        self.filter_button.setCheckable(True)

        # Set actions for tree widgets
        self.treewidget.previous_action = self.previous_action
        self.treewidget.next_action = self.next_action
        self.treewidget.filter_button = self.filter_button

        self.remote_treewidget.previous_action = self.previous_action
        self.remote_treewidget.next_action = self.next_action
        self.remote_treewidget.filter_button = self.filter_button

        # Setup widgets
        self.treewidget.setup()
        self.chdir(getcwd_or_home())

        # Menu
        menu = self.get_options_menu()

        hidden_action = self.get_action(DirViewActions.ToggleHiddenFiles)
        for item in [hidden_action, filters_action]:
            self.add_item_to_menu(
                item,
                menu=menu,
                section=ExplorerWidgetOptionsMenuSections.Common,
            )

        # Header actions
        size_column_action = self.get_action(DirViewActions.ToggleSizeColumn)
        type_column_action = self.get_action(DirViewActions.ToggleTypeColumn)
        date_column_action = self.get_action(DirViewActions.ToggleDateColumn)

        for item in [
            size_column_action,
            type_column_action,
            date_column_action,
        ]:
            self.add_item_to_menu(
                item,
                menu=menu,
                section=ExplorerWidgetOptionsMenuSections.Header,
            )

        single_click_action = self.get_action(DirViewActions.ToggleSingleClick)
        self.add_item_to_menu(
            single_click_action,
            menu=menu, section=ExplorerWidgetOptionsMenuSections.Files)

        # Toolbar
        toolbar = self.get_main_toolbar()
        for item in [
            self.previous_action,
            self.next_action,
            self.parent_action,
        ]:
            self.add_item_to_toolbar(
                item,
                toolbar=toolbar,
                section=ExplorerWidgetMainToolbarSections.Main,
            )

        for action in [
            self.remote_treewidget.upload_file_action,
            self.refresh_action,
            self.filter_button,
        ]:
            self.add_corner_widget(action, before=self._options_button)

    def update_actions(self):
        """Handle the update of actions of the plugin."""
        visible_remote_actions = (
            self.get_current_widget() == self.remote_treewidget
        )
        self.refresh_action.setVisible(visible_remote_actions)
        self.remote_treewidget.upload_file_action.setVisible(
            visible_remote_actions
        )

    # ---- Public API
    # ------------------------------------------------------------------------
    def chdir(self, directory, emit=True, server_id=None):
        """
        Set working directory.

        Parameters
        ----------
        directory: str
            Directory to set as working directory.
        emit: bool, optional
            Default is True.
        server_id: str, optional
            The server identification from where the directory is reachable.
            Default is None.
        """
        if not server_id:
            self.treewidget.chdir(directory, emit=emit)
            self.stackwidget.setCurrentWidget(self.treewidget)
        else:
            self.start_spinner()
            logger.debug(f"Request ls for {server_id} at {directory}")

            @AsyncDispatcher.QtSlot
            def remote_ls(future):
                self.remote_treewidget.chdir(
                    directory,
                    server_id=server_id,
                    emit=emit,
                    remote_files_manager=future.result()
                )

            self._plugin._get_remote_files_manager(server_id).connect(
                remote_ls
            )
            self.stackwidget.setCurrentWidget(self.remote_treewidget)

        self.update_actions()

    def get_current_folder(self):
        """Get current folder in the tree widget."""
        return self.stackwidget.currentWidget().get_current_folder()

    def set_current_folder(self, folder):
        """
        Set the current folder in the tree widget.

        Parameters
        ----------
        folder: str
            Folder path to set as current folder.
        """
        self.get_current_widget().set_current_folder(folder)

    def go_to_parent_directory(self):
        """Move to parent directory."""
        self.get_current_widget().go_to_parent_directory()

    def go_to_previous_directory(self):
        """Move to previous directory in history."""
        self.get_current_widget().go_to_previous_directory()

    def go_to_next_directory(self):
        """Move to next directory in history."""
        self.get_current_widget().go_to_next_directory()

    def refresh(self, new_path=None, force_current=False):
        """
        Refresh history.

        Parameters
        ----------
        new_path: str, optional
            Path to add to history. Default is None.
        force_current: bool, optional
            Default is True.
        """
        self.get_current_widget().refresh(new_path, force_current)

    def get_current_widget(self):
        """Get the current widget in stackwidget."""
        return self.stackwidget.currentWidget()

    @Slot()
    def edit_filter(self):
        """Edit name filters."""
        # Create Dialog
        dialog = QDialog(self)
        dialog.resize(500, 300)
        dialog.setWindowTitle(_('Edit filter settings'))

        # Create dialog contents
        description_label = QLabel(
            _(
                "Filter files by name, extension, or more using "
                '<a href="https://en.wikipedia.org/wiki/Glob_(programming)">'
                "glob patterns.</a> Please enter the glob patterns of the "
                "files you want to show, separated by commas."
            )
        )
        description_label.setOpenExternalLinks(True)
        description_label.setWordWrap(True)
        filters = QTextEdit(
            ", ".join(self.get_conf('name_filters')),
            parent=self
        )
        layout = QVBoxLayout()
        layout.addWidget(description_label)
        layout.addWidget(filters)

        def handle_ok():
            filter_text = filters.toPlainText()
            filter_text = [f.strip() for f in str(filter_text).split(',')]
            self.get_current_widget().set_name_filters(filter_text)
            dialog.accept()

        def handle_reset():
            self.get_current_widget().set_name_filters(NAME_FILTERS)
            filters.setPlainText(", ".join(self.get_conf('name_filters')))

        # Dialog buttons
        button_box = SpyderDialogButtonBox(
            QDialogButtonBox.Reset
            | QDialogButtonBox.Ok
            | QDialogButtonBox.Cancel
        )
        button_box.accepted.connect(handle_ok)
        button_box.rejected.connect(dialog.reject)
        button_box.button(QDialogButtonBox.Reset).clicked.connect(handle_reset)
        layout.addWidget(button_box)
        dialog.setLayout(layout)
        dialog.show()

    def change_filter_state(self):
        self.stackwidget.currentWidget().change_filter_state()

    def update_history(self, directory):
        """
        Update history with directory.

        Parameters
        ----------
        directory: str
            Path to add to history.
        """
        self.treewidget.update_history(directory)

    def reset_remote_treewidget(self, server_id):
        self.remote_treewidget.reset(server_id)

# =============================================================================
# Tests
# =============================================================================
class FileExplorerTest(QWidget):
    def __init__(self, directory=None, file_associations=None):

        self.CONF_SECTION = 'explorer'
        super().__init__()

        if directory is not None:
            self.directory = directory
        else:
            self.directory = osp.dirname(osp.abspath(__file__))

        file_associations = (
            {} if file_associations is None else file_associations
        )

        self.explorer = ExplorerWidget('explorer', self, parent=self)
        self.explorer.set_conf('file_associations', file_associations)
        self.explorer._setup()
        self.explorer.setup()
        self.label_dir = QLabel("<b>Open dir:</b>")
        self.label_file = QLabel("<b>Open file:</b>")
        self.label1 = QLabel()
        self.label_dir.setAlignment(Qt.AlignRight)
        self.label2 = QLabel()
        self.label_option = QLabel("<b>Option changed:</b>")
        self.label3 = QLabel()

        # Setup
        self.explorer.set_current_folder(self.directory)
        self.label_file.setAlignment(Qt.AlignRight)
        self.label_option.setAlignment(Qt.AlignRight)

        # Layout
        hlayout1 = QHBoxLayout()
        hlayout1.addWidget(self.label_file)
        hlayout1.addWidget(self.label1)

        hlayout2 = QHBoxLayout()
        hlayout2.addWidget(self.label_dir)
        hlayout2.addWidget(self.label2)

        hlayout3 = QHBoxLayout()
        hlayout3.addWidget(self.label_option)
        hlayout3.addWidget(self.label3)

        vlayout = QVBoxLayout()
        vlayout.addWidget(self.explorer)
        vlayout.addLayout(hlayout1)
        vlayout.addLayout(hlayout2)
        vlayout.addLayout(hlayout3)
        self.setLayout(vlayout)

        # Signals
        self.explorer.sig_dir_opened.connect(self.label2.setText)
        self.explorer.sig_dir_opened.connect(
            lambda: self.explorer.treewidget.refresh('..'))
        self.explorer.sig_open_file_requested.connect(self.label1.setText)


def test():
    from spyder.utils.qthelpers import qapplication
    app = qapplication()
    test = FileExplorerTest()
    test.resize(640, 480)
    test.show()
    app.exec_()


if __name__ == "__main__":
    test()



# -----------------------------------------------------------------------------
# Copyright (c) 2024- Spyder Project Contributors
#
# Released under the terms of the MIT License
# (see LICENSE.txt in the project root directory for details)
# -----------------------------------------------------------------------------

"""
Spyder dialog widgets.
"""

from __future__ import annotations

# Standard library imports
from typing import TYPE_CHECKING

# Third-party imports
from qtpy import QT6
from qtpy.QtCore import Qt
from qtpy.QtGui import QIcon
from qtpy.QtWidgets import QDialogButtonBox, QProxyStyle, QStyle

# Local imports
from spyder.utils.stylesheet import AppStyle

if TYPE_CHECKING:
    from qtpy.QtWidgets import QWidget


class _SpyderButtonsProxyStyle(QProxyStyle):
    """Style adjustments for :class:`SpyderDialogButtonBox`."""

    def styleHint(self, hint, option=None, widget=None, return_data=None):
        if hint == QStyle.SH_DialogButtonLayout:
            # Use the Windows buttons layout to have a uniform layout in all
            # platforms. We selected that layout because Windows is our most
            # popular platform.
            # Solution found in https://stackoverflow.com/a/35907926/438386
            if QT6:  # PySide6/PyQt6
                return QDialogButtonBox.ButtonLayout.WinLayout.value
            else:  # PySide2/PyQt5
                return int(QDialogButtonBox.ButtonLayout.WinLayout)

        return super().styleHint(hint, option, widget, return_data)


class SpyderDialogButtonBox(QDialogButtonBox):
    """Action button section (OK, Cancel, etc) widget for dialog boxes."""

    def __init__(
        self,
        buttons: QDialogButtonBox.StandardButton | None = None,
        orientation: Qt.Orientation = Qt.Horizontal,
        parent: QWidget | None = None,
    ):
        """
        Create a new group of dialog box buttons.

        Parameters
        ----------
        buttons : QDialogButtonBox.StandardButton | None, optional
            The standard buttons to include in the dialog box button group,
            such as "OK", "Cancel" or "Apply", passed as a single argument
            composed of Qt enum members bitwise or-ed (``|``) together.
            For example, ``QDialogButtonBox.Ok | QDialogButtonBox.Cancel``.
            If ``None``, the default, is an empty button box with no buttons.
        orientation : Qt.Orientation, optional
            Whether the button box is laid out horizontally, with buttons
            side by side, or vertically, with buttons on top of one another.
            By default, horizontal (``Qt.Horizontal``).
        parent : QWidget | None, optional
            The parent widget of this button box, ``None`` by default.

        Returns
        -------
        None
        """
        if buttons:
            super().__init__(buttons, orientation, parent)
        elif orientation:
            super().__init__(orientation=orientation, parent=parent)
        else:
            super().__init__(parent=parent)

        # Don't display icons on standard buttons. This is a problem on Linux
        button_constants = [
            QDialogButtonBox.Ok,
            QDialogButtonBox.Open,
            QDialogButtonBox.Save,
            QDialogButtonBox.Cancel,
            QDialogButtonBox.Close,
            QDialogButtonBox.Discard,
            QDialogButtonBox.Apply,
            QDialogButtonBox.Reset,
            QDialogButtonBox.RestoreDefaults,
            QDialogButtonBox.Help,
            QDialogButtonBox.SaveAll,
            QDialogButtonBox.Yes,
            QDialogButtonBox.YesToAll,
            QDialogButtonBox.No,
            QDialogButtonBox.NoToAll,
            QDialogButtonBox.Abort,
            QDialogButtonBox.Retry,
            QDialogButtonBox.Ignore,
        ]

        for constant in button_constants:
            button = self.button(constant)
            if button is not None:
                button.setIcon(QIcon())

        # Set a reasonable spacing between buttons. This is a problem on Mac
        self.layout().setSpacing(2 * AppStyle.MarginSize)

        # Set style
        style = _SpyderButtonsProxyStyle(None)
        style.setParent(self)
        self.setStyle(style)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Collapsible widget to hide and show child widgets."""

import qstylizer.style
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QPushButton
from superqt import QCollapsible

from spyder.utils.icon_manager import ima
from spyder.utils.palette import SpyderPalette
from spyder.utils.stylesheet import AppStyle


class CollapsibleWidget(QCollapsible):
    """Collapsible widget to hide and show child widgets."""

    def __init__(self, parent=None, title=""):
        super().__init__(title=title, parent=parent)

        # Align widget to the left to text before or after it (don't know why
        # this is necessary).
        self.layout().setContentsMargins(5, 0, 0, 0)

        # Remove spacing between toggle button and contents area
        self.layout().setSpacing(0)

        # Set icons
        self.setCollapsedIcon(ima.icon("collapsed"))
        self.setExpandedIcon(ima.icon("expanded"))

        # To change the style only of these widgets
        self._toggle_btn.setObjectName("collapsible-toggle")
        self.content().setObjectName("collapsible-content")

        # Add padding to the inside content
        self.content().layout().setContentsMargins(
            *((AppStyle.InnerContentPadding,) * 4)
        )

        # Set stylesheet
        self._css = self._generate_stylesheet()
        self.setStyleSheet(self._css.toString())

        # Signals
        self.toggled.connect(self._on_toggled)

        # Set our properties for the toggle button
        self._set_toggle_btn_properties()

    def set_content_bottom_margin(self, bottom_margin):
        """Set bottom margin of the content area to `bottom_margin`."""
        margins = self.content().layout().contentsMargins()
        margins.setBottom(bottom_margin)
        self.content().layout().setContentsMargins(margins)

    def set_content_right_margin(self, right_margin):
        """Set right margin of the content area to `right_margin`."""
        margins = self.content().layout().contentsMargins()
        margins.setRight(right_margin)
        self.content().layout().setContentsMargins(margins)

    def _generate_stylesheet(self):
        """Generate base stylesheet for this widget."""
        css = qstylizer.style.StyleSheet()

        # --- Style for the header button
        css["QPushButton#collapsible-toggle"].setValues(
            # Increase padding (the default one is too small).
            padding=f"{2 * AppStyle.MarginSize}px",
            # Make it a bit different from a default QPushButton to not drag
            # the same amount of attention to it.
            backgroundColor=SpyderPalette.COLOR_BACKGROUND_3
        )

        # Make hover color match the change of background color above
        css["QPushButton#collapsible-toggle:hover"].setValues(
            backgroundColor=SpyderPalette.COLOR_BACKGROUND_4,
        )

        # --- Style for the contents area
        css["QWidget#collapsible-content"].setValues(
            # Remove top border to make it appear attached to the header button
            borderTop="0px",
            # Add border to the other edges
            border=f'1px solid {SpyderPalette.COLOR_BACKGROUND_4}',
            # Add border radius to the bottom to make it match the style of our
            # other widgets.
            borderBottomLeftRadius=f'{SpyderPalette.SIZE_BORDER_RADIUS}',
            borderBottomRightRadius=f'{SpyderPalette.SIZE_BORDER_RADIUS}',
        )

        return css

    def _on_toggled(self, state):
        """Adjustments when the button is toggled."""
        if state:
            # Remove bottom rounded borders from the header when the widget is
            # expanded.
            self._css["QPushButton#collapsible-toggle"].setValues(
                borderBottomLeftRadius='0px',
                borderBottomRightRadius='0px',
            )
        else:
            # Restore bottom rounded borders to the header when the widget is
            # collapsed.
            self._css["QPushButton#collapsible-toggle"].setValues(
                borderBottomLeftRadius=f'{SpyderPalette.SIZE_BORDER_RADIUS}',
                borderBottomRightRadius=f'{SpyderPalette.SIZE_BORDER_RADIUS}',
            )

        self.setStyleSheet(self._css.toString())

    def _set_toggle_btn_properties(self):
        """Set properties for the toogle button."""

        def enter_event(event):
            self.setCursor(Qt.PointingHandCursor)
            super(QPushButton, self._toggle_btn).enterEvent(event)

        def leave_event(event):
            self.setCursor(Qt.ArrowCursor)
            super(QPushButton, self._toggle_btn).leaveEvent(event)

        self.toggleButton().enterEvent = enter_event
        self.toggleButton().leaveEvent = leave_event



# -----------------------------------------------------------------------------
# Copyright (c) 2021- Spyder Project Contributors
#
# Released under the terms of the MIT License
# (see LICENSE.txt in the project root directory for details)
# -----------------------------------------------------------------------------

"""
Spyder API plugin registration mixins.

.. deprecated:: 6.2

    This module will be moved to a private :mod:`!spyder.api._mixins` module
    and become a deprecated alias, raising a :exc:`DeprecationWarning`,
    that will be removed in Spyder 7.0.

    It is a private implementation detail of the plugin decorators in
    :mod:`spyder.api.plugin_registration.decorators`, and wasn't designed
    or intended to be used directly by external code. Plugins access its
    functionality through the :class:`~spyder.api.plugins.SpyderPluginV2`
    class instead.
"""

from __future__ import annotations

# Standard library imports
import logging

from spyder.api.exceptions import SpyderAPIError
from spyder.api.plugins import Plugins

logger = logging.getLogger(__name__)


class SpyderPluginObserver:
    """
    Mixin to receive and respond to changes in Spyder plugin availability.

    This mixin enables a class to receive notifications when a plugin
    is available, by registering methods using the
    :func:`~spyder.api.plugin_registration.decorators.on_plugin_available`
    decorator. When any of the requested plugins is ready, the corresponding
    registered method is called.

    Normally inherited and initialized automatically through
    :class:`~spyder.api.plugins.SpyderPluginV2` rather than used directly.

    .. caution::

        This mixin will only operate over the plugins listed under the
        :attr:`~spyder.api.plugins.SpyderPluginV2.REQUIRES` or
        :attr:`~spyder.api.plugins.SpyderPluginV2.OPTIONAL` class constants
        of the class inheriting the mixin.
    """

    def __init__(self) -> None:
        """
        Set up the plugin listeners for any decorated methods of the class.

        Called automatically by
        :meth:`SpyderPluginV2.__init__() <spyder.api.plugins.SpyderPluginV2.__init__>`.

        Returns
        -------
        None

        Raises
        ------
        SpyderAPIError
            If trying to watch a plugin that is not listed in the plugin class'
            :attr:`~spyder.api.plugins.SpyderPluginV2.REQUIRES`
            or :attr:`~spyder.api.plugins.SpyderPluginV2.OPTIONAL` class constants.
        """
        self._plugin_listeners = {}
        self._plugin_teardown_listeners = {}
        for method_name in dir(self):
            method = getattr(self, method_name, None)
            if hasattr(method, "_plugin_listen"):
                plugin_listen = method._plugin_listen

                # Check if plugin is listed among REQUIRES and OPTIONAL.
                # Note: We can't do this validation for the Layout plugin
                # because it depends on all plugins through the Plugins.All
                # wildcard.
                if self.NAME != Plugins.Layout and (
                    plugin_listen not in self.REQUIRES + self.OPTIONAL
                ):
                    raise SpyderAPIError(
                        f"Method {method_name} of {self} is trying to watch "
                        f"plugin {plugin_listen}, but that plugin is not "
                        f"listed in REQUIRES nor OPTIONAL."
                    )

                logger.debug(
                    f"Method {method_name} is watching plugin {plugin_listen}"
                )
                self._plugin_listeners[plugin_listen] = method_name

            if hasattr(method, "_plugin_teardown"):
                plugin_teardown = method._plugin_teardown

                # Check if plugin is listed among REQUIRES and OPTIONAL.
                # Note: We can't do this validation for the Layout plugin
                # because it depends on all plugins through the Plugins.All
                # wildcard.
                if self.NAME != Plugins.Layout and (
                    plugin_teardown not in self.REQUIRES + self.OPTIONAL
                ):
                    raise SpyderAPIError(
                        f"Method {method_name} of {self} is trying to watch "
                        f"plugin {plugin_teardown}, but that plugin is not "
                        f"listed in REQUIRES nor OPTIONAL."
                    )

                logger.debug(
                    f"Method {method_name} will handle plugin "
                    f"teardown for {plugin_teardown}"
                )
                self._plugin_teardown_listeners[plugin_teardown] = method_name

    def _on_plugin_available(self, plugin: str) -> None:
        """
        Handle plugin availability and redirect it to plugin-specific handlers.

        Parameters
        ----------
        plugin: str
            Name of the plugin that was notified as available.

        Returns
        -------
        None
        """
        # Call plugin specific handler
        if plugin in self._plugin_listeners:
            method_name = self._plugin_listeners[plugin]
            method = getattr(self, method_name)
            logger.debug(f"Calling {method}")
            method()

        # Call global plugin handler
        if "__all" in self._plugin_listeners:
            method_name = self._plugin_listeners["__all"]
            method = getattr(self, method_name)
            method(plugin)

    def _on_plugin_teardown(self, plugin: str) -> None:
        """
        Handle plugin teardown and redirect it to plugin-specific handlers.

        Parameters
        ----------
        plugin: str
            Name of the plugin that is going through its teardown process.

        Returns
        -------
        None
        """
        # Call plugin specific handler
        if plugin in self._plugin_teardown_listeners:
            method_name = self._plugin_teardown_listeners[plugin]
            method = getattr(self, method_name)
            logger.debug(f"Calling {method}")
            method()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2009- Spyder Project Contributors
#
# Distributed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Remote Client Plugin.
"""

# Standard library imports
from __future__ import annotations
import logging
import typing

# Third-party imports
from qtpy.QtCore import Signal

# Local imports
from spyder.api.asyncdispatcher import AsyncDispatcher
from spyder.api.plugin_registration.decorators import (
    on_plugin_available,
    on_plugin_teardown,
)
from spyder.api.plugins import OptionalPlugins, Plugins, SpyderPluginV2
from spyder.api.translations import _
from spyder.plugins.mainmenu.api import (
    ApplicationMenus,
    ToolsMenuSections,
)
from spyder.plugins.remoteclient.api import (
    RemoteClientActions,
)
from spyder.plugins.remoteclient.api.manager.base import (
    SpyderRemoteAPIManagerBase,
)
from spyder.plugins.remoteclient.api.manager.jupyterhub import (
    SpyderRemoteJupyterHubAPIManager,
)
from spyder.plugins.remoteclient.api.manager.ssh import (
    SpyderRemoteSSHAPIManager,
)
from spyder.plugins.remoteclient.api.protocol import (
    ClientType,
    JupyterHubClientOptions,
    SSHClientOptions,
    ConnectionStatus,
)
from spyder.plugins.remoteclient.api.modules.file_services import (
    SpyderRemoteFileServicesAPI,
)
from spyder.plugins.remoteclient.api.modules.environ import (
    SpyderRemoteEnvironAPI,
)
from spyder.plugins.remoteclient.widgets import AuthenticationMethod
from spyder.plugins.remoteclient.widgets.container import RemoteClientContainer

if typing.TYPE_CHECKING:
    from spyder.plugins.remoteclient.api.modules.base import (
        SpyderBaseJupyterAPIType,
    )


_logger = logging.getLogger(__name__)


class RemoteClient(SpyderPluginV2):
    """
    Remote client plugin.
    """

    NAME = "remoteclient"
    OPTIONAL = [Plugins.MainMenu, OptionalPlugins.EnvManager]
    CONF_SECTION = NAME
    CONTAINER_CLASS = RemoteClientContainer
    CONF_FILE = False

    CONF_SECTION_SERVERS = "servers"

    # ---- Signals
    sig_server_stopped = Signal(str)
    sig_server_renamed = Signal(str)
    sig_server_changed = Signal()
    sig_client_message_logged = Signal(dict)

    sig_connection_established = Signal(str)
    sig_connection_lost = Signal(str)
    sig_connection_status_changed = Signal(dict)

    sig_version_mismatch = Signal(str, str)

    # For remote envs
    sig_create_env_requested = Signal(str, str, str, list)
    """
    Signal to request creating a Python environment in the remote machine.

    Parameters
    ----------
    config_id: str
        Machine identifier.
    env_name: str
        Environment name.
    python_version: str
        Python version of the environment.
    packages_list: list[str]
        List of packages.
    """

    sig_import_env_requested = Signal(str, str, str)
    """
    Signal to request importing a Python environment in the remote machine.

    Parameters
    ----------
    config_id: str
        Machine identifier.
    import_file_path: str
        Path to the file that will be used to import the environment.
    env_name: str
        Environment name.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._remote_clients: dict[str, SpyderRemoteAPIManagerBase] = {}

    # ---- SpyderPluginV2 API
    # -------------------------------------------------------------------------
    @staticmethod
    def get_name():
        return _("Remote client")

    @staticmethod
    def get_description():
        return _("Connect to remote machines to run code on them.")

    @classmethod
    def get_icon(cls):
        return cls.create_icon("remote_server")

    def on_initialize(self):
        self._reset_status()
        container = self.get_container()

        # Container signals
        container.sig_start_server_requested.connect(self.start_remote_server)
        container.sig_stop_server_requested.connect(self.stop_remote_server)
        container.sig_stop_server_requested.connect(self.sig_server_stopped)
        container.sig_server_renamed.connect(self.sig_server_renamed)
        container.sig_server_changed.connect(self.sig_server_changed)
        container.sig_server_updated.connect(self.load_client_from_id)

        # Plugin signals
        self.sig_connection_status_changed.connect(
            container.sig_connection_status_changed
        )
        self.sig_client_message_logged.connect(
            container.sig_client_message_logged
        )
        self.sig_version_mismatch.connect(container.on_server_version_mismatch)

    def on_first_registration(self):
        pass

    def on_close(self, cancellable=True):
        """Stops remote server and close any opened connection."""
        while self._remote_clients:
            client = self._remote_clients.popitem()[1]
            AsyncDispatcher(
                loop="asyncssh",
                early_return=False,
            )(client.close)()

    @on_plugin_available(plugin=Plugins.MainMenu)
    def on_mainmenu_available(self):
        mainmenu = self.get_plugin(Plugins.MainMenu)

        action = self.get_action(RemoteClientActions.ManageConnections)
        mainmenu.add_item_to_application_menu(
            action,
            menu_id=ApplicationMenus.Tools,
            section=ToolsMenuSections.Managers,
            before_section=ToolsMenuSections.Preferences,
        )

    @on_plugin_teardown(plugin=Plugins.MainMenu)
    def on_mainmenu_teardown(self):
        mainmenu = self.get_plugin(Plugins.MainMenu)

        mainmenu.remove_item_from_application_menu(
            RemoteClientActions.ManageConnections,
            menu_id=ApplicationMenus.Tools,
        )

    @on_plugin_available(plugin=OptionalPlugins.EnvManager)
    def on_envmanager_available(self):
        envmanager = self.get_plugin(OptionalPlugins.EnvManager)
        envmanager.sig_set_default_kernel_spec_requested.connect(
            self.set_default_kernel_spec
        )

    @on_plugin_teardown(plugin=OptionalPlugins.EnvManager)
    def on_envmanager_teardown(self):
        envmanager = self.get_plugin(OptionalPlugins.EnvManager)
        envmanager.sig_set_default_kernel_spec_requested.disconnect(
            self.set_default_kernel_spec
        )

    # ---- Public API
    # -------------------------------------------------------------------------
    # --- Remote Server Methods
    def get_remote_server(self, config_id):
        """Get remote server."""
        if config_id in self._remote_clients:
            return self._remote_clients[config_id]

    @AsyncDispatcher(loop="asyncssh")
    async def _install_remote_server(self, config_id):
        """Install remote server."""
        if config_id in self._remote_clients:
            client = self._remote_clients[config_id]
            await client.connect_and_install_remote_server()

    def start_remote_server(self, config_id):
        """Start remote server."""
        if config_id not in self._remote_clients:
            self.load_client_from_id(config_id)

        @AsyncDispatcher(loop="asyncssh")
        async def _start_client():
            client = self._remote_clients[config_id]
            await client.connect_and_ensure_server()
        _start_client()

    @AsyncDispatcher(loop="asyncssh")
    async def stop_remote_server(self, config_id):
        """Stop remote server."""
        if client := self._remote_clients.pop(config_id, None):
            await client.close()

    @AsyncDispatcher(loop="asyncssh")
    async def ensure_remote_server(self, config_id):
        """Ensure remote server is running and installed."""
        if config_id in self._remote_clients:
            client = self._remote_clients[config_id]
            await client.connect_and_ensure_server()

    def restart_remote_server(self, config_id):
        """Restart remote server."""
        self.stop_remote_server(config_id)
        self.start_remote_server(config_id)

    # --- Configuration Methods
    def load_client_from_id(self, config_id):
        """Load remote server from configuration id."""
        client_type = self.get_conf(
            f"{config_id}/client_type", default=ClientType.SSH
        )
        if client_type == ClientType.SSH:
            options = self._load_ssh_client_options(config_id)
            self.load_ssh_client(config_id, options)
        elif client_type == ClientType.JupyterHub:
            options = self._load_jupyterhub_client_options(config_id)
            self.load_jupyterhub_client(config_id, options)
        else:
            msg = (
                f"Unknown client type '{client_type}' for server "
                f"'{config_id}'. Please check your configuration."
            )
            raise ValueError(msg)

    def load_ssh_client(self, config_id: str, options: SSHClientOptions):
        """Load SSH remote server."""
        if config_id in self._remote_clients:
            self._remote_clients[config_id].options.update(options)
        else:
            self._remote_clients[config_id] = SpyderRemoteSSHAPIManager(
                config_id, options, _plugin=self
            )

        _logger.debug(
            "Remote SSH client for '%s' loaded with options: %s",
            config_id,
            options,
        )

    def load_jupyterhub_client(
        self, config_id: str, options: JupyterHubClientOptions,
    ):
        """Load JupyterHub remote server."""
        if config_id in self._remote_clients:
            self._remote_clients[config_id].options.update(options)
        else:
            self._remote_clients[config_id] = SpyderRemoteJupyterHubAPIManager(
                config_id, options, _plugin=self
            )

        _logger.debug(
            "Remote JupyterHub client for '%s' loaded with options: %s",
            config_id,
            options,
        )

    def _load_jupyterhub_client_options(self, config_id):
        """Load JupyterHub remote server configuration."""
        options = JupyterHubClientOptions(
            url=self.get_conf(f"{config_id}/url"),
            token=self.get_conf(f"{config_id}/token", secure=True)
        )

        return options

    def _load_ssh_client_options(self, config_id):
        """Load remote server configuration."""
        auth_method = self.get_conf(f"{config_id}/auth_method")

        options = SSHClientOptions(
            host=self.get_conf(
                f"{config_id}/{auth_method}/address"
            ),
            port=self.get_conf(f"{config_id}/{auth_method}/port"),
            username=self.get_conf(
                f"{config_id}/{auth_method}/username"
            ),
        )

        if client_keys := self.get_conf(f"{config_id}/keyfile", ""):
            options["client_keys"] = [client_keys]
            if passphrase := self.get_conf(
                f"{config_id}/passphrase", "", secure=True
            ):
                options["passphrase"] = passphrase
        elif config := self.get_conf(f"{config_id}/configfile"):
            options["config"] = config
        else:
            # Password is mandatory in this case
            password = self.get_conf(f"{config_id}/password", secure=True)
            options["password"] = password

        # Default for now
        options["platform"] = "linux"

        # Don't require the host to be known to connect to it
        options["known_hosts"] = None

        return SSHClientOptions(**options)

    def get_loaded_servers(self):
        """Get configured remote servers."""
        return self._remote_clients.keys()

    def get_config_ids(self):
        """Get configured remote servers ids."""
        return self.get_conf(self.CONF_SECTION_SERVERS, {}).keys()

    def get_server_name(self, config_id):
        """Get configured remote server name."""
        client_type = self.get_conf(
            f"{config_id}/client_type", default=ClientType.SSH
        )
        if client_type == ClientType.SSH:
            auth_method = self.get_conf(f"{config_id}/auth_method")
            return self.get_conf(f"{config_id}/{auth_method}/name")
        if client_type == ClientType.JupyterHub:
            return self.get_conf(
                f"{config_id}/{AuthenticationMethod.JupyterHub}/name"
            )

        msg = (
            f"Unknown client type '{client_type}' for server '{config_id}'. "
            f"Please check your configuration."
        )
        raise ValueError(msg)

    def set_default_kernel_spec(self, name: str, server_id: str):
        """Set the default kernel spec for a remote server.

        This method sets the default kernel spec for a remote server
        identified by `server_id`. The kernel spec is used to determine
        which kernel spec is used when starting a new console on that server.

        Parameters
        ----------
        name: str
            Name of the kernel spec to set as default.
        server_id: str
            Identifier of the remote server for which to set
            the default kernel spec.

        Raises
        ------
        ValueError
            If the server_id does not exist in the configuration.
        """
        configs = self.get_conf(self.CONF_SECTION_SERVERS, {})
        if server_id not in configs:
            msg = f"Configuration for remote server '{server_id}' not found."
            raise ValueError(msg)

        configs[server_id]["default_kernel_spec"] = name

        self.set_conf(self.CONF_SECTION_SERVERS, configs)

        if server_id in self._remote_clients:
            server_options = self._remote_clients[server_id].options
            server_options["default_kernel_spec"] = name

    # --- API Methods
    @staticmethod
    def register_api(kclass: typing.Type[SpyderBaseJupyterAPIType]):
        """
        Register Remote Client API.

        This method is used to register a new API class that will be used to
        interact with the remote server. It can be used as a decorator.

        Parameters
        ----------
        kclass: Type[SpyderBaseJupyterAPI]
            Class to be registered.

        Returns
        -------
        Type[SpyderBaseJupyterAPI]
            Class that was registered.
        """
        return SpyderRemoteAPIManagerBase.register_api(kclass)

    def get_api(
        self, config_id: str, api: str | typing.Type[SpyderBaseJupyterAPIType]
    ):
        """
        Get the API for a remote server.

        Get the registered API class for a given remote server.

        Parameters
        ----------
        config_id: str
            Configuration id of the remote server.
        api: str | Type[SpyderBaseJupyterAPI]
            API class to be retrieved.

        Returns
        -------
        SpyderBaseJupyterAPI
            API class instance.
        """
        if config_id not in self._remote_clients:
            self.load_client_from_id(config_id)

        client = self._remote_clients[config_id]

        return client.get_api(api)

    def get_file_api(self, config_id):
        """Get file API."""
        if config_id not in self._remote_clients:
            self.load_client_from_id(config_id)

        client = self._remote_clients[config_id]

        return client.get_api(SpyderRemoteFileServicesAPI)

    def get_environ_api(self, config_id):
        """Get environment API."""
        if config_id not in self._remote_clients:
            self.load_client_from_id(config_id)

        client = self._remote_clients[config_id]

        return client.get_api(SpyderRemoteEnvironAPI)

    def get_jupyter_api(self, config_id):
        """Get Jupyter API."""
        if config_id not in self._remote_clients:
            self.load_client_from_id(config_id)

        client = self._remote_clients[config_id]

        return client.get_jupyter_api()

    # ---- Private API
    # -------------------------------------------------------------------------
    # --- Remote Server Kernel Methods
    def _reset_status(self):
        """Reset status of servers."""
        for config_id in self.get_config_ids():
            self.set_conf(f"{config_id}/status", ConnectionStatus.Inactive)
            self.set_conf(f"{config_id}/status_message", "")

    @AsyncDispatcher(loop="asyncssh")
    async def _abort_connection(self, config_id: str):
        if config_id not in self._remote_clients:
            return

        client = self._remote_clients[config_id]
        await client.abort_connection()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Kernel spec for Spyder kernels
"""

# Standard library imports
import logging
import os
import os.path as osp

try:
    from win32api import GetShortPathName
except Exception:
    GetShortPathName = None

# Third party imports
from jupyter_client.kernelspec import KernelSpec
from packaging.version import parse
from requests.structures import CaseInsensitiveDict
from spyder_kernels.utils.pythonenv import (
    get_conda_env_path,
    get_pixi_manifest_path_and_env_name,
    is_conda_env,
    is_pixi_env,
)

# Local imports
from spyder.api.config.mixins import SpyderConfigurationAccessor
from spyder.api.translations import _
from spyder.config.base import (
    get_safe_mode,
    is_conda_based_app,
    running_in_ci,
    running_under_pytest
)
from spyder.plugins.ipythonconsole import (
    SPYDER_KERNELS_CONDA, SPYDER_KERNELS_PIP, SPYDER_KERNELS_VERSION,
    SpyderKernelError)
from spyder.utils.conda import conda_version, find_conda, find_pixi
from spyder.utils.environ import clean_env
from spyder.utils.misc import get_python_executable
from spyder.utils.programs import (
    get_module_version,
    get_temp_dir,
    is_python_interpreter,
    is_module_installed,
)

# Constants
HERE = os.path.abspath(os.path.dirname(__file__))
logger = logging.getLogger(__name__)
ERROR_SPYDER_KERNEL_INSTALLED = _(
    "The Python environment or installation whose interpreter is located at"
    "<pre>"
    "    <tt>{0}</tt>"
    "</pre>"
    "doesn't have <tt>spyder-kernels</tt> version <tt>{1}</tt> installed. "
    "Without this module and specific version is not possible for Spyder to "
    "create a console for you.<br><br>"
    "You can install it by activating your environment (if necessary) and "
    "then running in a system terminal:"
    "<pre>"
    "    <tt>{2}</tt>"
    "</pre>"
    "or"
    "<pre>"
    "    <tt>{3}</tt>"
    "</pre>")


def has_spyder_kernels(pyexec):
    """Check if env has spyder kernels."""
    if is_module_installed(
        'spyder_kernels',
        version=SPYDER_KERNELS_VERSION,
        interpreter=pyexec
    ):
        return True

    # Dev versions of Spyder-kernels are acceptable
    try:
        return "dev0" in get_module_version('spyder_kernels', pyexec)
    except Exception:
        return False


HERE = osp.dirname(os.path.realpath(__file__))


class SpyderKernelSpec(KernelSpec, SpyderConfigurationAccessor):
    """Kernel spec for Spyder kernels"""

    CONF_SECTION = 'ipython_console'

    def __init__(self, path_to_custom_interpreter=None, **kwargs):
        super().__init__(**kwargs)
        self.path_to_custom_interpreter = path_to_custom_interpreter
        self.display_name = 'Python 3 (Spyder)'
        self.language = 'python3'
        self.resource_dir = ''
        self._env_vars = {}

    @property
    def argv(self):
        """Command to start kernels"""
        # Python interpreter used to start kernels
        if (
            self.get_conf('default', section='main_interpreter')
            and not self.path_to_custom_interpreter
        ):
            pyexec = get_python_executable()
        else:
            pyexec = self.get_conf('executable', section='main_interpreter')
            if self.path_to_custom_interpreter:
                pyexec = self.path_to_custom_interpreter
            if not has_spyder_kernels(pyexec):
                raise SpyderKernelError(
                    ERROR_SPYDER_KERNEL_INSTALLED.format(
                        pyexec,
                        SPYDER_KERNELS_VERSION,
                        SPYDER_KERNELS_CONDA,
                        SPYDER_KERNELS_PIP
                    )
                )
                return
            if not is_python_interpreter(pyexec):
                pyexec = get_python_executable()
                self.set_conf('executable', '', section='main_interpreter')
                self.set_conf('default', True, section='main_interpreter')
                self.set_conf('custom', False, section='main_interpreter')

        # Command used to start kernels
        kernel_cmd = []

        if is_pixi_env(pyexec=pyexec):
            pixi_exe = find_pixi()

            if not pixi_exe:
                raise SpyderKernelError(
                    _(
                        "Spyder couldn't find pixi in your system to activate "
                        "the kernel's environment. Please add the directory "
                        "where the pixi executable is located to your PATH "
                        "environment variable for it to be detected."
                    )
                )
            pixi_manifest, pixi_env = get_pixi_manifest_path_and_env_name(
                pyexec,
            )
            kernel_cmd.extend([
                pixi_exe,
                'run',
                '--environment',
                pixi_env,
                '--manifest-path',
                pixi_manifest,
            ])

        elif is_conda_env(pyexec=pyexec):
            # If executable is a conda environment, use "run" subcommand to
            # activate it and run spyder-kernels.
            conda_exe = find_conda()

            if not conda_exe:
                conda_exe = (
                    self.get_conf("conda_path", section="main_interpreter")
                    if self.get_conf(
                        "custom_conda", section="main_interpreter"
                    )
                    else None
                )
                if not conda_exe:
                    # Raise error since we were unable to determine the path to
                    # the conda executable (e.g when Anaconda/Miniconda was
                    # installed in a non-standard location).
                    # See spyder-ide/spyder#23595
                    not_found_exe_message = _(
                        "Spyder couldn't find Conda, Mamba or Micromamba on your "
                        "system to activate the kernel's environment.<br><br>"
                        "Please set the path for one of their executables in "
                        "<tt>Preferences > Python interpreter > Conda "
                        "executable</tt>"
                    )
                    raise SpyderKernelError(not_found_exe_message)

            # Get conda/mamba/micromamba version to perform some checks
            conda_exe_version = conda_version(conda_executable=conda_exe)

            # Base command
            kernel_cmd.extend([
                conda_exe,
                'run',
                '--prefix',
                get_conda_env_path(pyexec)
            ])

            # We need to use these flags to prevent conda_exe from capturing
            # the kernel process stdout/stderr streams. That way we'll be able
            # to show them in Spyder.
            if "micromamba" in osp.basename(conda_exe) or (
                # Fixes spyder-ide/spyder#24513
                "mamba" in osp.basename(conda_exe)
                and conda_exe_version >= parse("2.0")
            ):
                kernel_cmd.extend(['--attach', '""'])
            elif (
                "mamba" in osp.basename(conda_exe)
                and conda_exe_version < parse("2.0")
            ) or (
                "conda" in osp.basename(conda_exe)
                and conda_exe_version >= parse("4.9")
            ):
                # Note: We use --no-capture-output instead of --live-stream
                # here because it works for older Conda versions (conda>=4.9).
                kernel_cmd.append('--no-capture-output')
            else:
                # Raise error since an unsupported conda version is being used
                # (conda<4.9).
                # See spyder-ide/spyder#22554
                raise SpyderKernelError(
                    _(
                        "The detected version of Conda is too old and not "
                        "supported by Spyder. The minimum supported version is "
                        "4.9 and currently you have {conda_version}.<br><br>."
                        "<b>Note</b>: You need to restart Spyder after "
                        "updating Conda for the change to take effect."
                    ).format(conda_version=conda_exe_version)
                )

        kernel_cmd.extend([
            pyexec,
            # This is necessary to avoid a spurious message on Windows.
            # Fixes spyder-ide/spyder#20800.
            '-Xfrozen_modules=off',
            '-m', 'spyder_kernels.console',
            '-f', '{connection_file}'
        ])

        logger.info('Kernel command: {}'.format(kernel_cmd))

        return kernel_cmd

    @property
    def env(self):
        """Environment variables for kernels"""
        return self._env_vars

    @env.setter
    def env(self, env_vars):
        """Setter for environment variables for kernels"""
        env_vars = dict(env_vars)
        if os.name == "nt":
            # Expand variables in case some are unexpanded.
            # Fixes spyder-ide/spyder#25275
            env_vars = {k: osp.expandvars(v) for k, v in env_vars.items()}

            # Use case insensitive dictionary
            env_vars = CaseInsensitiveDict(env_vars)

            # HKCU path must be appended to HKLM path
            path = os.getenv("path", "").split(";")  # HKLM Path
            path.extend(env_vars.get("path", "").split(";"))  # HKCU Path
            path = ";".join([p for p in path if p])  # Stringify
            env_vars["PATH"] = path

            # Use short path name for temporary directories in case path has
            # spaces. See spyder-ide/spyder#25403
            if GetShortPathName is not None:
                if "tmp" in env_vars:
                    env_vars["tmp"] = GetShortPathName(env_vars["tmp"])
                if "temp" in env_vars:
                    env_vars["temp"] = GetShortPathName(env_vars["temp"])

        # User variables supersede system variables
        for k, v in os.environ.items():
            env_vars.setdefault(k, v)

        default_interpreter = self.get_conf(
            'default', section='main_interpreter'
        )

        # List of modules to exclude from our UMR
        umr_namelist = self.get_conf(
            'umr/namelist', section='main_interpreter')

        # Get TMPDIR value, if available
        tmpdir_var = env_vars.get("TMPDIR", "")

        # --- Adding to environment variables

        # Environment variables that we need to pass to the kernel
        env_vars.update({
            'SPY_EXTERNAL_INTERPRETER': (not default_interpreter
                                         or self.path_to_custom_interpreter),
            'SPY_UMR_ENABLED': self.get_conf(
                'umr/enabled', section='main_interpreter'),
            'SPY_UMR_VERBOSE': self.get_conf(
                'umr/verbose', section='main_interpreter'),
            'SPY_UMR_NAMELIST': ','.join(umr_namelist),
            'SPY_AUTOCALL_O': self.get_conf('autocall'),
            'SPY_GREEDY_O': self.get_conf('greedy_completer'),
            'SPY_JEDI_O': self.get_conf('jedi_completer'),
            'SPY_TESTING': running_under_pytest() or get_safe_mode(),
            'SPY_HIDE_CMD': self.get_conf('hide_cmd_windows'),
            # This env var avoids polluting the OS default temp directory with
            # files generated by `conda run`. It's restored/removed in the
            # kernel after initialization.
            "TMPDIR": get_temp_dir(),
            # This is necessary to restore TMPDIR in the kernel, if it exists
            "SPY_TMPDIR": tmpdir_var,
        })

        # This is necessary so that the kernel checks the Spyder pid for
        # crashes and gets killed automatically when that happens.
        # Fixes spyder-ide/spyder#22414
        if not (running_in_ci() and os.name == "nt"):
            env_vars["SPY_PARENT_PID"] = str(os.getpid())

        # App considerations
        # ??? Do we need this?
        if is_conda_based_app() and default_interpreter:
            # See spyder-ide/spyder#16927
            # See spyder-ide/spyder#16828
            # See spyder-ide/spyder#17552
            env_vars['PYDEVD_DISABLE_FILE_VALIDATION'] = 1

        # --- Removing from envrionment variables

        env_vars.pop('PYTEST_CURRENT_TEST', None)

        # Avoid IPython adding the virtualenv on which Spyder is running
        # to the kernel sys.path
        env_vars.pop('VIRTUAL_ENV', None)

        # Do not pass PYTHONPATH to kernels directly, spyder-ide/spyder#13519
        env_vars.pop('PYTHONPATH', None)

        # Remove this variable because it prevents starting kernels for
        # external interpreters when present.
        # Fixes spyder-ide/spyder#13252
        env_vars.pop('PYTHONEXECUTABLE', None)

        # Making all env_vars strings, ensure cast as dict
        self._env_vars = clean_env(dict(env_vars))



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Pythopath container."""

from collections import OrderedDict
import logging
import os
import os.path as osp

from qtpy.QtCore import Signal

from spyder.api.translations import _
from spyder.api.widgets.main_container import PluginMainContainer
from spyder.config.base import get_conf_path
from spyder.plugins.pythonpath.widgets.pathmanager import PathManager


# Logging
logger = logging.getLogger(__name__)


# ---- Constants
# -----------------------------------------------------------------------------
class PythonpathActions:
    Manager = "manager_action"


# ---- Container
# -----------------------------------------------------------------------------
class PythonpathContainer(PluginMainContainer):

    sig_pythonpath_changed = Signal(object, bool)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # ---- PluginMainContainer API
    # -------------------------------------------------------------------------
    def setup(self):
        # Migrate to new config options if necessary
        if not self.get_conf("config_options_migrated", False):
            self._migrate_to_config_options()

        # This attribute is only used to detect changes and after initializing
        # here should only be set in update_active_project_path.
        self._project_path = OrderedDict()

        # These attributes are only used to detect changes and after
        # initializing here should only be set in _save_paths.
        self._user_paths = OrderedDict(self.get_conf('user_paths'))
        self._system_paths = self.get_conf('system_paths')
        self._prioritize = self.get_conf('prioritize')
        self._spyder_pythonpath = self.get_conf('spyder_pythonpath')

        # Path manager dialog
        self.path_manager_dialog = PathManager(parent=self, sync=True)
        self.path_manager_dialog.sig_path_changed.connect(
            self._save_paths
        )
        self.path_manager_dialog.redirect_stdio.connect(
            self.sig_redirect_stdio_requested)

        # Path manager action
        self.path_manager_action = self.create_action(
            PythonpathActions.Manager,
            _("PYTHONPATH manager"),
            icon=self.create_icon('pythonpath'),
            triggered=self.show_path_manager
        )

    def update_actions(self):
        pass

    # ---- Public API
    # -------------------------------------------------------------------------
    def update_active_project_path(self, path):
        """
        Update active project path.

        _project_path is set in this method and nowhere else.
        """
        # _project_path should be reset whenever it is updated.
        self._project_path = OrderedDict()
        if path is None:
            logger.debug("Update Spyder PYTHONPATH because project was closed")
        else:
            logger.debug(f"Add project paths to Spyder PYTHONPATH: {path}")
            self._project_path.update({path: True})

        self._save_paths()

    def show_path_manager(self):
        """
        Show path manager dialog.
        """
        # Do not update paths if widget is already open,
        # see spyder-ide/spyder#20808.
        if not self.path_manager_dialog.isVisible():
            self.path_manager_dialog.update_paths(
                project_path=self._project_path,
                user_paths=self._user_paths,
                system_paths=self._system_paths,
                prioritize=self._prioritize
            )

        # Show and give it focus
        self.path_manager_dialog.show()
        self.path_manager_dialog.activateWindow()
        self.path_manager_dialog.raise_()
        self.path_manager_dialog.setFocus()

    def get_spyder_pythonpath(self):
        """Return active Spyder PYTHONPATH as a list of paths."""
        # Desired behavior is project_path | user_paths | system_paths, but
        # Python 3.8 does not support | operator for OrderedDict.
        all_paths = OrderedDict(reversed(self._system_paths.items()))
        all_paths.update(reversed(self._user_paths.items()))
        all_paths.update(reversed(self._project_path.items()))
        all_paths = OrderedDict(reversed(all_paths.items()))

        return [p for p, v in all_paths.items() if v]

    # ---- Private API
    # -------------------------------------------------------------------------
    def _save_paths(self, user_paths=None, system_paths=None, prioritize=None):
        """
        Save user and system path dictionaries and prioritize to config.

        Parameters
        ----------
        user_paths: OrderedDict
            Paths set by the user.
        system_paths: OrderedDict
            Paths set in the PYTHONPATH environment variable.
        prioritize: bool
            Whether paths should be prepended (True) or appended (False) to
            sys.path.

        Notes
        -----
        - Each dictionary key is a path and the value is the active state.
        - sig_pythonpath_changed is emitted from this method, and nowhere else,
        on condition that _spyder_pythonpath changed.

        """
        assert isinstance(user_paths, (type(None), OrderedDict))
        assert isinstance(system_paths, (type(None), OrderedDict))
        assert isinstance(prioritize, (type(None), bool))

        emit = False

        # Don't set options unless necessary
        if user_paths is not None and user_paths != self._user_paths:
            logger.debug(f"Saving user paths: {user_paths}")
            self.set_conf('user_paths', dict(user_paths))
            self._user_paths = user_paths

        if system_paths is not None and system_paths != self._system_paths:
            logger.debug(f"Saving system paths: {system_paths}")
            self.set_conf('system_paths', dict(system_paths))
            self._system_paths = system_paths

        if prioritize is not None and prioritize != self._prioritize:
            logger.debug(f"Saving prioritize: {prioritize}")
            self.set_conf('prioritize', prioritize)
            self._prioritize = prioritize
            emit = True

        spyder_pythonpath = self.get_spyder_pythonpath()
        if spyder_pythonpath != self._spyder_pythonpath:
            logger.debug(f"Saving Spyder pythonpath: {spyder_pythonpath}")
            self.set_conf('spyder_pythonpath', spyder_pythonpath)
            self._spyder_pythonpath = spyder_pythonpath
            emit = True

        # Only emit signal if spyder_pythonpath or prioritize changed
        if emit:
            self.sig_pythonpath_changed.emit(
                self._spyder_pythonpath, self._prioritize
            )

    def _migrate_to_config_options(self):
        """
        Migrate paths saved in the `path` and `not_active_path` files located
        in our config directory to our config system.

        # TODO: Remove for Spyder 7
        """
        path_file = get_conf_path('path')
        not_active_path_file = get_conf_path('not_active_path')
        config_path = self.get_conf('path', ())
        config_not_active_path = self.get_conf('not_active_path', ())
        system_path = self.get_conf('system_path', ())

        path = []
        not_active_path = []

        # Get path from file
        if osp.isfile(path_file):
            with open(path_file, 'r', encoding='utf-8') as f:
                path = f.read().splitlines()
            try:
                os.remove(path_file)
            except OSError:
                pass

        # Get inactive paths from file
        if osp.isfile(not_active_path_file):
            with open(not_active_path_file, 'r', encoding='utf-8') as f:
                not_active_path = f.read().splitlines()
            try:
                os.remove(not_active_path_file)
            except OSError:
                pass

        # Get path from config; supersedes paths from file
        if config_path:
            path = config_path

        # Get inactive path from config; supersedes paths from file
        if config_not_active_path is not None:
            not_active_path = config_not_active_path

        # Get system path
        system_paths = {}
        if system_path:
            system_paths = {p: p not in not_active_path for p in system_path}

        # path config has all user and system paths; only want user paths
        user_paths = {
            p: p not in not_active_path for p in path if p not in system_path
        }


        # Update the configuration
        self.set_conf('user_paths', user_paths)
        self.set_conf('system_paths', system_paths)

        # Do not migrate again
        self.set_conf("config_options_migrated", True)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Application Plugin.
"""

# Standard library imports
import os
import os.path as osp
import subprocess
import sys
from typing import Dict, Optional, Tuple

# Third party imports
from qtpy.QtCore import Slot

# Local imports
from spyder.api.plugins import Plugins, SpyderDockablePlugin, SpyderPluginV2
from spyder.api.translations import _
from spyder.api.plugin_registration.decorators import (
    on_plugin_available, on_plugin_teardown)
from spyder.api.plugin_registration.registry import PLUGIN_REGISTRY
from spyder.api.widgets.menus import SpyderMenu, MENU_SEPARATOR
from spyder.app import SHORTCUT_EXE
from spyder.config.base import (get_module_path, get_debug_level,
                                running_under_pytest)
from spyder.plugins.application.confpage import ApplicationConfigPage
from spyder.plugins.application.container import (
    ApplicationActions, ApplicationContainer, ApplicationPluginMenus)
from spyder.plugins.console.api import ConsoleActions
from spyder.plugins.editor.api.actions import EditorWidgetActions
from spyder.plugins.mainmenu.api import (
    ApplicationMenus,
    EditMenuSections,
    FileMenuSections,
    HelpMenuSections,
    SearchMenuSections,
    ToolsMenuSections,
)
from spyder.plugins.toolbar.api import ApplicationToolbars
from spyder.utils.misc import getcwd_or_home
from spyder.utils.qthelpers import add_actions


class Application(SpyderPluginV2):
    NAME = 'application'
    REQUIRES = [Plugins.Console, Plugins.Preferences]
    OPTIONAL = [
        Plugins.Editor,
        Plugins.Help,
        Plugins.IPythonConsole,
        Plugins.MainMenu,
        Plugins.StatusBar,
        Plugins.Shortcuts,  # Needed to display the app context menu
        Plugins.Toolbar,
        Plugins.UpdateManager,  # Required in the confpage
    ]
    CONTAINER_CLASS = ApplicationContainer
    CONF_SECTION = 'main'
    CONF_FILE = False
    CONF_WIDGET_CLASS = ApplicationConfigPage
    CAN_BE_DISABLED = False

    def __init__(self, parent, configuration=None):
        super().__init__(parent, configuration)
        self.focused_plugin: Optional[SpyderDockablePlugin] = None
        self.file_action_enabled: Dict[Tuple[str, str], bool] = {}
        self.edit_action_enabled: Dict[Tuple[str, str], bool] = {}
        self.search_action_enabled: Dict[Tuple[str, str], bool] = {}

    @staticmethod
    def get_name():
        return _('Application')

    @classmethod
    def get_icon(cls):
        return cls.create_icon('genprefs')

    @staticmethod
    def get_description():
        return _('Provide main application base actions.')

    def on_initialize(self):
        container = self.get_container()
        container.sig_report_issue_requested.connect(self.report_issue)
        container.sig_new_file_requested.connect(self.create_new_file)
        container.sig_open_file_in_plugin_requested.connect(
            self.open_file_in_plugin
        )
        container.sig_open_file_using_dialog_requested.connect(
            self.open_file_using_dialog
        )
        container.sig_open_last_closed_requested.connect(
            self.open_last_closed_file
        )
        container.sig_save_file_requested.connect(self.save_file)
        container.sig_save_all_requested.connect(self.save_all)
        container.sig_save_file_as_requested.connect(self.save_file_as)
        container.sig_save_copy_as_requested.connect(self.save_copy_as)
        container.sig_revert_file_requested.connect(self.revert_file)
        container.sig_close_file_requested.connect(self.close_file)
        container.sig_close_all_requested.connect(self.close_all)
        container.sig_undo_requested.connect(self.undo)
        container.sig_redo_requested.connect(self.redo)
        container.sig_cut_requested.connect(self.cut)
        container.sig_copy_requested.connect(self.copy)
        container.sig_paste_requested.connect(self.paste)
        container.sig_select_all_requested.connect(self.select_all)
        container.sig_find_requested.connect(self.find)
        container.sig_find_next_requested.connect(self.find_next)
        container.sig_find_previous_requested.connect(self.find_previous)
        container.sig_replace_requested.connect(self.replace)
        container.set_window(self._window)
        self.sig_focused_plugin_changed.connect(self._update_focused_plugin)

    # ---- Plugin initialization
    # -------------------------------------------------------------------------
    @on_plugin_available(plugin=Plugins.IPythonConsole)
    def on_ipythonconsole_available(self):
        if self.is_plugin_available(Plugins.MainMenu):
            self._populate_help_menu()

    @on_plugin_available(plugin=Plugins.Console)
    def on_console_available(self):
        if self.is_plugin_available(Plugins.MainMenu):
            self.report_action.setVisible(True)

    @on_plugin_available(plugin=Plugins.Preferences)
    def on_preferences_available(self):
        # Register conf page
        preferences = self.get_plugin(Plugins.Preferences)
        preferences.register_plugin_preferences(self)

    @on_plugin_available(plugin=Plugins.MainMenu)
    def on_main_menu_available(self):
        self._populate_file_menu()
        self._populate_edit_menu()
        self._populate_search_menu()
        self._populate_tools_menu()

        if self.is_plugin_enabled(Plugins.IPythonConsole):
            if self.is_plugin_available(Plugins.IPythonConsole):
                self._populate_help_menu()
        else:
            self._populate_help_menu()

        if not self.is_plugin_available(Plugins.Console):
            self.report_action.setVisible(False)

    @on_plugin_available(plugin=Plugins.Editor)
    def on_editor_available(self):
        editor = self.get_plugin(Plugins.Editor)
        self.get_container().sig_load_log_file.connect(editor.load)

    @on_plugin_available(plugin=Plugins.StatusBar)
    def on_statusbar_available(self):
        statusbar = self.get_plugin(Plugins.StatusBar)
        inapp_appeal_status = self.get_container().inapp_appeal_status
        statusbar.add_status_widget(inapp_appeal_status)

    @on_plugin_available(plugin=Plugins.Toolbar)
    def on_toolbar_available(self):
        container = self.get_container()
        toolbar = self.get_plugin(Plugins.Toolbar)
        for action in [
            container.new_action,
            container.open_action,
            container.save_action,
            container.save_all_action
        ]:
            toolbar.add_item_to_application_toolbar(
                action,
                toolbar_id=ApplicationToolbars.File,
                before=EditorWidgetActions.NewCell
            )

    @on_plugin_available(plugin=Plugins.UpdateManager)
    def on_update_manager_available(self):
        update_manager = self.get_plugin(Plugins.UpdateManager)
        update_manager.sig_update_performed.connect(self._on_spyder_update)

    # ---- Plugin Teardown
    # -------------------------------------------------------------------------
    @on_plugin_teardown(plugin=Plugins.Preferences)
    def on_preferences_teardown(self):
        preferences = self.get_plugin(Plugins.Preferences)
        preferences.deregister_plugin_preferences(self)

    @on_plugin_teardown(plugin=Plugins.Editor)
    def on_editor_teardown(self):
        editor = self.get_plugin(Plugins.Editor)
        self.get_container().sig_load_log_file.disconnect(editor.load)

    @on_plugin_teardown(plugin=Plugins.Console)
    def on_console_teardown(self):
        if self.is_plugin_available(Plugins.MainMenu):
            self.report_action.setVisible(False)

    @on_plugin_teardown(plugin=Plugins.MainMenu)
    def on_main_menu_teardown(self):
        self._depopulate_file_menu()
        self._depopulate_edit_menu()
        self._depopulate_search_menu()
        self._depopulate_tools_menu()
        self._depopulate_help_menu()
        self.report_action.setVisible(False)

    @on_plugin_teardown(plugin=Plugins.StatusBar)
    def on_statusbar_teardown(self):
        statusbar = self.get_plugin(Plugins.StatusBar)
        inapp_appeal_status = self.get_container().inapp_appeal_status
        statusbar.remove_status_widget(inapp_appeal_status.ID)

    @on_plugin_teardown(plugin=Plugins.Toolbar)
    def on_toolbar_teardown(self):
        toolbar = self.get_plugin(Plugins.Toolbar)
        for action in [
            ApplicationActions.NewFile,
            ApplicationActions.OpenFile,
            ApplicationActions.SaveFile,
            ApplicationActions.SaveAll
        ]:
            toolbar.remove_item_from_application_toolbar(
                action,
                toolbar_id=ApplicationToolbars.File
            )

    @on_plugin_teardown(plugin=Plugins.UpdateManager)
    def on_update_manager_teardowm(self):
        update_manager = self.get_plugin(Plugins.UpdateManager)
        update_manager.sig_update_performed.disconnect(self._on_spyder_update)

    def on_close(self, _unused=True):
        self.get_container().on_close()

    def on_mainwindow_visible(self):
        """Actions after the mainwindow in visible."""
        container = self.get_container()

        # Show dialog with missing dependencies
        if not running_under_pytest():
            container.compute_dependencies()

        # Handle DPI scale and window changes to show a restart message.
        # Don't activate this functionality on macOS because it's being
        # triggered in the wrong situations.
        # See spyder-ide/spyder#11846
        if not sys.platform == 'darwin':
            window = self._window.windowHandle()
            window.screenChanged.connect(container.handle_new_screen)
            screen = self._window.windowHandle().screen()
            container.current_dpi = screen.logicalDotsPerInch()
            screen.logicalDotsPerInchChanged.connect(
                container.show_dpi_change_message)

        # Show appeal the fifth and 25th time Spyder starts
        spyder_runs = self.get_conf("spyder_runs_for_appeal", default=1)
        if spyder_runs in [5, 25]:
            container.inapp_appeal_status.show_appeal()

            # Increase counting in one to not get stuck at this point.
            # Fixes spyder-ide/spyder#22457
            self.set_conf("spyder_runs_for_appeal", spyder_runs + 1)
        else:
            if spyder_runs < 25:
                self.set_conf("spyder_runs_for_appeal", spyder_runs + 1)

    # ---- Private API
    # ------------------------------------------------------------------------
    def _populate_file_menu(self):
        container = self.get_container()
        mainmenu = self.get_plugin(Plugins.MainMenu)

        # New section
        mainmenu.add_item_to_application_menu(
            container.new_action,
            menu_id=ApplicationMenus.File,
            section=FileMenuSections.New,
            before_section=FileMenuSections.Open
        )

        # Open section
        open_actions = [
            container.open_action,
            container.open_last_closed_action,
            container.recent_files_menu,
        ]
        for open_action in open_actions:
            mainmenu.add_item_to_application_menu(
                open_action,
                menu_id=ApplicationMenus.File,
                section=FileMenuSections.Open,
                before_section=FileMenuSections.Save
            )

        # Save section
        save_actions = [
            container.save_action,
            container.save_all_action,
            container.save_as_action,
            container.save_copy_as_action,
            container.revert_action
        ]
        for save_action in save_actions:
            mainmenu.add_item_to_application_menu(
                save_action,
                menu_id=ApplicationMenus.File,
                section=FileMenuSections.Save,
                before_section=FileMenuSections.Print
            )

        # Close section
        close_actions = [
            container.close_file_action,
            container.close_all_action
        ]
        for close_action in close_actions:
            mainmenu.add_item_to_application_menu(
                close_action,
                menu_id=ApplicationMenus.File,
                section=FileMenuSections.Close,
                before_section=FileMenuSections.Restart
            )

        # Restart section
        mainmenu.add_item_to_application_menu(
            self.restart_action,
            menu_id=ApplicationMenus.File,
            section=FileMenuSections.Restart
        )
        mainmenu.add_item_to_application_menu(
            self.restart_debug_action,
            menu_id=ApplicationMenus.File,
            section=FileMenuSections.Restart
        )

    def _populate_edit_menu(self):
        container = self.get_container()
        mainmenu = self.get_plugin(Plugins.MainMenu)

        # Undo/Redo section
        for action in [container.undo_action, container.redo_action]:
            mainmenu.add_item_to_application_menu(
                action,
                menu_id=ApplicationMenus.Edit,
                section=EditMenuSections.UndoRedo,
                before_section=EditMenuSections.Editor
            )

        # Edit section
        edit_actions = [
            container.cut_action,
            container.copy_action,
            container.paste_action,
            container.select_all_action
        ]
        for edit_action in edit_actions:
            mainmenu.add_item_to_application_menu(
                edit_action,
                menu_id=ApplicationMenus.Edit,
                section=EditMenuSections.Copy,
                before_section=EditMenuSections.Editor
            )

    def _populate_search_menu(self):
        container = self.get_container()
        mainmenu = self.get_plugin(Plugins.MainMenu)

        # Find section
        for action in [
            container.find_action,
            container.find_next_action,
            container.find_previous_action,
            container.replace_action,
        ]:
            mainmenu.add_item_to_application_menu(
                action,
                menu_id=ApplicationMenus.Search,
                section=SearchMenuSections.FindInText,
                before_section=SearchMenuSections.Cursor,
            )

    def _populate_tools_menu(self):
        """Add base actions and menus to the Tools menu."""
        mainmenu = self.get_plugin(Plugins.MainMenu)
        mainmenu.add_item_to_application_menu(
            self.user_env_action,
            menu_id=ApplicationMenus.Tools,
            section=ToolsMenuSections.Managers,
            before_section=ToolsMenuSections.Preferences,
        )

        if get_debug_level() >= 2:
            mainmenu.add_item_to_application_menu(
                self.debug_logs_menu,
                menu_id=ApplicationMenus.Help,
                section=HelpMenuSections.Support,
                before_section=HelpMenuSections.About,
            )

    def _populate_help_menu(self):
        """Add base actions and menus to the Help menu."""
        self._populate_help_menu_documentation_section()
        self._populate_help_menu_support_section()
        self._populate_help_menu_about_section()

    def _populate_help_menu_documentation_section(self):
        """Add base Spyder documentation actions to the Help main menu."""
        mainmenu = self.get_plugin(Plugins.MainMenu)
        ipythonconsole = self.get_plugin(Plugins.IPythonConsole)
        ipython_documentation_submenu = None

        if ipythonconsole:
            from spyder.plugins.ipythonconsole.api import (
                IPythonConsoleWidgetMenus
            )
            ipython_documentation_submenu = (
                IPythonConsoleWidgetMenus.Documentation
            )
        for documentation_action in [
                self.documentation_action, self.video_action]:
            mainmenu.add_item_to_application_menu(
                documentation_action,
                menu_id=ApplicationMenus.Help,
                section=HelpMenuSections.ExternalDocumentation,
                before=ipython_documentation_submenu,
                before_section=HelpMenuSections.Support,
            )

    def _populate_help_menu_support_section(self):
        """Add Spyder base support actions to the Help main menu."""
        mainmenu = self.get_plugin(Plugins.MainMenu)
        for support_action in [
            self.trouble_action,
            self.support_group_action,
            self.dependencies_action,
            self.report_action,
        ]:
            mainmenu.add_item_to_application_menu(
                support_action,
                menu_id=ApplicationMenus.Help,
                section=HelpMenuSections.Support,
                before_section=HelpMenuSections.About,
            )

    def _populate_help_menu_about_section(self):
        """Create Spyder base about actions."""
        mainmenu = self.get_plugin(Plugins.MainMenu)
        for action in [
            self.get_action(ApplicationActions.ShowChangelogAction),
            self.get_action(ApplicationActions.HelpSpyderAction),
            self.about_action,
        ]:
            mainmenu.add_item_to_application_menu(
                action,
                menu_id=ApplicationMenus.Help,
                section=HelpMenuSections.About,
            )

    @property
    def _window(self):
        return self.main.window()

    def _depopulate_help_menu(self):
        self._depopulate_help_menu_documentation_section()
        self._depopulate_help_menu_support_section()
        self._depopulate_help_menu_about_section()

    def _depopulate_help_menu_documentation_section(self):
        mainmenu = self.get_plugin(Plugins.MainMenu)
        for documentation_action in [
                ApplicationActions.SpyderDocumentationAction,
                ApplicationActions.SpyderDocumentationVideoAction]:
            mainmenu.remove_item_from_application_menu(
                documentation_action,
                menu_id=ApplicationMenus.Help)

    def _depopulate_help_menu_support_section(self):
        """Remove Spyder base support actions from the Help main menu."""
        mainmenu = self.get_plugin(Plugins.MainMenu)
        for support_action in [
                ApplicationActions.SpyderTroubleshootingAction,
                ConsoleActions.SpyderReportAction,
                ApplicationActions.SpyderDependenciesAction,
                ApplicationActions.SpyderSupportAction]:
            mainmenu.remove_item_from_application_menu(
                support_action,
                menu_id=ApplicationMenus.Help)

    def _depopulate_help_menu_about_section(self):
        mainmenu = self.get_plugin(Plugins.MainMenu)
        mainmenu.remove_item_from_application_menu(
            ApplicationActions.SpyderAbout,
            menu_id=ApplicationMenus.Help)

    def _depopulate_file_menu(self):
        container = self.get_container()
        mainmenu = self.get_plugin(Plugins.MainMenu)
        for action_id in [
            ApplicationActions.NewFile,
            ApplicationActions.OpenFile,
            ApplicationActions.OpenLastClosed,
            container.recent_file_menu,
            ApplicationActions.SaveFile,
            ApplicationActions.SaveAll,
            ApplicationActions.SaveAs,
            ApplicationActions.SaveCopyAs,
            ApplicationActions.RevertFile,
            ApplicationActions.CloseFile,
            ApplicationActions.CloseAll,
            ApplicationActions.SpyderRestart,
            ApplicationActions.SpyderRestartDebug
        ]:
            mainmenu.remove_item_from_application_menu(
                action_id,
                menu_id=ApplicationMenus.File)

    def _depopulate_edit_menu(self):
        mainmenu = self.get_plugin(Plugins.MainMenu)
        for action_id in [
            ApplicationActions.Undo,
            ApplicationActions.Redo,
            ApplicationActions.Cut,
            ApplicationActions.Copy,
            ApplicationActions.Paste,
            ApplicationActions.SelectAll,
        ]:
            mainmenu.remove_item_from_application_menu(
                action_id,
                menu_id=ApplicationMenus.File)

    def _depopulate_search_menu(self):
        mainmenu = self.get_plugin(Plugins.MainMenu)
        for action_id in [
            ApplicationActions.FindText,
            ApplicationActions.FindNext,
            ApplicationActions.FindPrevious,
            ApplicationActions.ReplaceText,
        ]:
            mainmenu.remove_item_from_application_menu(
                action_id,
                menu_id=ApplicationMenus.Search
            )

    def _depopulate_tools_menu(self):
        """Add base actions and menus to the Tools menu."""
        mainmenu = self.get_plugin(Plugins.MainMenu)
        mainmenu.remove_item_from_application_menu(
            ApplicationActions.SpyderUserEnvVariables,
            menu_id=ApplicationMenus.Tools)

        if get_debug_level() >= 2:
            mainmenu.remove_item_from_application_menu(
                ApplicationPluginMenus.DebugLogsMenu,
                menu_id=ApplicationMenus.Tools)

    def _update_focused_plugin(
        self, plugin: Optional[SpyderDockablePlugin]
    ) -> None:
        """
        Update which plugin has currently focus.

        This function is called if another plugin gets keyboard focus.
        """
        self.focused_plugin = plugin
        self._update_file_actions()
        self._update_edit_actions()
        self._update_search_actions()

    def _update_file_actions(self) -> None:
        """
        Update which file actions are enabled.

        File actions are enabled depending on whether the plugin that would
        currently process the file action has enabled it or not.
        """
        plugin = self.focused_plugin
        if not plugin or not getattr(plugin, 'CAN_HANDLE_FILE_ACTIONS', False):
            plugin = self.get_plugin(Plugins.Editor, error=False)
        if plugin:
            for action_name in [
                ApplicationActions.NewFile,
                ApplicationActions.OpenLastClosed,
                ApplicationActions.SaveFile,
                ApplicationActions.SaveAll,
                ApplicationActions.SaveAs,
                ApplicationActions.SaveCopyAs,
                ApplicationActions.RevertFile,
                ApplicationActions.CloseFile,
                ApplicationActions.CloseAll,
            ]:
                action = self.get_action(action_name)
                key = (plugin.NAME, action_name)
                state = self.file_action_enabled.get(key, True)
                action.setEnabled(state)

    def _update_edit_actions(self) -> None:
        """
        Update which edit actions are enabled.

        Edit actions are enabled depending on whether the plugin that would
        currently process the edit action has enabled it or not.
        """
        plugin = self.focused_plugin
        if not plugin or not getattr(plugin, 'CAN_HANDLE_EDIT_ACTIONS', False):
            plugin = self.get_plugin(Plugins.Editor, error=False)
        if plugin:
            for action_name in [
                ApplicationActions.Undo,
                ApplicationActions.Redo,
                ApplicationActions.Cut,
                ApplicationActions.Copy,
                ApplicationActions.Paste,
                ApplicationActions.SelectAll,
            ]:
                action = self.get_action(action_name)
                key = (plugin.NAME, action_name)
                state = self.edit_action_enabled.get(key, True)
                action.setEnabled(state)

    def _update_search_actions(self) -> None:
        """
        Update which search actions are enabled.

        Search actions are enabled depending on whether the plugin that would
        currently process the search action has enabled it or not.
        """
        plugin = self.focused_plugin
        if not plugin or not getattr(plugin, 'CAN_HANDLE_SEARCH_ACTIONS', False):
            plugin = self.get_plugin(Plugins.Editor, error=False)
        if plugin:
            for action_name in [
                ApplicationActions.FindText,
                ApplicationActions.FindNext,
                ApplicationActions.FindPrevious,
                ApplicationActions.ReplaceText,
            ]:
                action = self.get_action(action_name)
                key = (plugin.NAME, action_name)
                state = self.search_action_enabled.get(key, True)
                action.setEnabled(state)

    def _on_spyder_update(self):
        container = self.get_container()
        container.inapp_appeal_status.show_changelog()

    # ---- Public API
    # ------------------------------------------------------------------------
    def get_application_context_menu(self, parent=None):
        """
        Return menu with the actions to be shown by the Spyder context menu.
        """
        tutorial_action = None
        shortcuts_action = None

        help_plugin = self.get_plugin(Plugins.Help)
        shortcuts = self.get_plugin(Plugins.Shortcuts)
        menu = SpyderMenu(parent=parent)
        actions = [self.documentation_action]
        # Help actions
        if help_plugin:
            from spyder.plugins.help.plugin import HelpActions
            tutorial_action = help_plugin.get_action(
                HelpActions.ShowSpyderTutorialAction)
            actions += [tutorial_action]
        # Shortcuts actions
        if shortcuts:
            from spyder.plugins.shortcuts.plugin import ShortcutActions
            shortcuts_action = shortcuts.get_action(
                ShortcutActions.ShortcutSummaryAction)
            actions.append(shortcuts_action)
        # Application actions
        actions += [MENU_SEPARATOR, self.about_action]

        add_actions(menu, actions)

        return menu

    def report_issue(self):
        if self.is_plugin_available(Plugins.Console):
            console = self.get_plugin(Plugins.Console)
            console.report_issue()

    def apply_settings(self):
        """Apply applications settings."""
        self._main.apply_settings()

    @Slot()
    def restart(self, reset=False, close_immediately=False):
        """
        Quit and Restart Spyder application.

        If reset True it allows to reset spyder on restart.
        """
        # Get console plugin reference to call the quit action
        console = self.get_plugin(Plugins.Console)

        # Get start path to use in restart script
        spyder_start_directory = get_module_path('spyder')
        restart_script = osp.join(spyder_start_directory, 'app', 'restart.py')

        # Get any initial argument passed when spyder was started
        # Note: Variables defined in bootstrap.py and spyder/app/start.py
        env = os.environ.copy()
        bootstrap_args = env.pop('SPYDER_BOOTSTRAP_ARGS', None)
        spyder_args = env.pop('SPYDER_ARGS')

        # Get current process and python running spyder
        pid = os.getpid()
        python = SHORTCUT_EXE or sys.executable

        # Check if started with bootstrap.py
        if bootstrap_args is not None:
            spyder_args = bootstrap_args
            is_bootstrap = True
        else:
            is_bootstrap = False

        # Pass variables as environment variables (str) to restarter subprocess
        env['SPYDER_ARGS'] = spyder_args
        env['SPYDER_PID'] = str(pid)
        env['SPYDER_IS_BOOTSTRAP'] = str(is_bootstrap)

        # Build the command and popen arguments depending on the OS
        if os.name == 'nt':
            # Hide flashing command prompt
            startupinfo = subprocess.STARTUPINFO()
            startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
            shell = False
        else:
            startupinfo = None
            shell = True

        command = '"{0}" "{1}"'
        command = command.format(python, restart_script)

        try:
            if self.main.closing(True, close_immediately=close_immediately):
                subprocess.Popen(command, shell=shell, env=env,
                                 startupinfo=startupinfo)
                console.quit()
        except Exception as error:
            # If there is an error with subprocess, Spyder should not quit and
            # the error can be inspected in the internal console
            print(error)  # spyder: test-skip
            print(command)  # spyder: test-skip

    def create_new_file(self) -> None:
        """
        Create new file in a suitable plugin.

        If the plugin that currently has focus, has its
        `CAN_HANDLE_FILE_ACTIONS` attribute set to `True`, then create a new
        file in that plugin. Otherwise, create a new file in the Editor plugin.
        """
        plugin = self.focused_plugin
        if plugin and getattr(plugin, 'CAN_HANDLE_FILE_ACTIONS', False):
            plugin.create_new_file()
        elif self.is_plugin_available(Plugins.Editor):
            editor = self.get_plugin(Plugins.Editor)
            editor.new()

    def open_file_using_dialog(self) -> None:
        """
        Show Open File dialog and open the selected file.

        Try asking the plugin that currently has focus for the name of the
        displayed file and whether it is a temporary file. If that does not
        work, ask the Editor plugin.
        """
        plugin = self.focused_plugin
        if plugin:
            filename = plugin.get_current_filename()
        else:
            filename = None

        if filename is None and self.is_plugin_available(Plugins.Editor):
            plugin = self.get_plugin(Plugins.Editor)
            filename = plugin.get_current_filename()

        if filename is not None and not plugin.current_file_is_temporary():
            basedir = osp.dirname(filename)
        else:
            basedir = getcwd_or_home()

        self.get_container().open_file_using_dialog(filename, basedir)

    def open_file_in_plugin(self, filename: str) -> None:
        """
        Open given file in a suitable plugin.

        Go through all plugins and open the file in the first plugin that
        registered the extension of the given file name. If none is found,
        then open the file in the Editor plugin.
        """
        ext = osp.splitext(filename)[1]
        for plugin_name in PLUGIN_REGISTRY:
            if PLUGIN_REGISTRY.is_plugin_available(plugin_name):
                plugin = PLUGIN_REGISTRY.get_plugin(plugin_name)
                if (
                    isinstance(plugin, SpyderDockablePlugin)
                    and ext in plugin.FILE_EXTENSIONS
                ):
                    plugin.switch_to_plugin()
                    plugin.open_file(filename)
                    return

        if self.is_plugin_available(Plugins.Editor):
            editor = self.get_plugin(Plugins.Editor)
            editor.load(filename)

    def open_last_closed_file(self) -> None:
        """
        Open the last closed file again.

        If the plugin that currently has focus, has its
        `CAN_HANDLE_FILE_ACTIONS` attribute set to `True`, then open the
        last closed file in that plugin. Otherwise, open the last closed file
        in the Editor plugin.
        """
        plugin = self.focused_plugin
        if plugin and getattr(plugin, 'CAN_HANDLE_FILE_ACTIONS', False):
            plugin.open_last_closed_file()
        elif self.is_plugin_available(Plugins.Editor):
            editor = self.get_plugin(Plugins.Editor)
            editor.open_last_closed()

    def add_recent_file(self, fname: str) -> None:
        """
        Add file to list of recent files.

        This function adds the given file name to the list of recent files,
        which is used in the `File > Open recent` menu. The function ensures
        that the list has no duplicates and it is no longer than the maximum
        length.
        """
        self.get_container().add_recent_file(fname)

    def save_file(self) -> None:
        """
        Save current file.

        If the plugin that currently has focus, has its
        `CAN_HANDLE_FILE_ACTIONS` attribute set to `True`, then save the
        current file in that plugin. Otherwise, save the current file in the
        Editor plugin.
        """
        plugin = self.focused_plugin
        if plugin and getattr(plugin, 'CAN_HANDLE_FILE_ACTIONS', False):
            plugin.save_file()
        elif self.is_plugin_available(Plugins.Editor):
            editor = self.get_plugin(Plugins.Editor)
            editor.save()

    def save_file_as(self) -> None:
        """
        Save current file under a different name.

        If the plugin that currently has focus, has its
        `CAN_HANDLE_FILE_ACTIONS` attribute set to `True`, then save the
        current file in that plugin under a different name. Otherwise, save
        the current file in the Editor plugin under a different name.
        """
        plugin = self.focused_plugin
        if plugin and getattr(plugin, 'CAN_HANDLE_FILE_ACTIONS', False):
            plugin.save_file_as()
        elif self.is_plugin_available(Plugins.Editor):
            editor = self.get_plugin(Plugins.Editor)
            editor.save_as()

    def save_copy_as(self) -> None:
        """
        Save copy of current file under a different name.

        If the plugin that currently has focus, has its
        `CAN_HANDLE_FILE_ACTIONS` attribute set to `True`, then save a copy of
        the current file in that plugin under a different name. Otherwise, save
        a copy of the current file in the Editor plugin under a different name.
        """
        plugin = self.focused_plugin
        if plugin and getattr(plugin, 'CAN_HANDLE_FILE_ACTIONS', False):
            plugin.save_copy_as()
        elif self.is_plugin_available(Plugins.Editor):
            editor = self.get_plugin(Plugins.Editor)
            editor.save_copy_as()

    def save_all(self) -> None:
        """
        Save all files.

        If the plugin that currently has focus, has its
        `CAN_HANDLE_FILE_ACTIONS` attribute set to `True`, then save all files
        in that plugin. Otherwise, save all files in the Editor plugin.
        """
        plugin = self.focused_plugin
        if plugin and getattr(plugin, 'CAN_HANDLE_FILE_ACTIONS', False):
            plugin.save_all()
        elif self.is_plugin_available(Plugins.Editor):
            editor = self.get_plugin(Plugins.Editor)
            editor.save_all()

    def revert_file(self) -> None:
        """
        Revert current file to the version on disk.

        If the plugin that currently has focus, has its
        `CAN_HANDLE_FILE_ACTIONS` attribute set to `True`, then revert the
        current file in that plugin to the version stored on disk. Otherwise,
        revert the current file in the Editor plugin.
        """
        plugin = self.focused_plugin
        if plugin and getattr(plugin, 'CAN_HANDLE_FILE_ACTIONS', False):
            plugin.revert_file()
        elif self.is_plugin_available(Plugins.Editor):
            editor = self.get_plugin(Plugins.Editor)
            editor.revert_file()

    def close_file(self) -> None:
        """
        Close the current file.

        If the plugin that currently has focus, has its
        `CAN_HANDLE_FILE_ACTIONS` attribute set to `True`, then close the
        current file in that plugin. Otherwise, close the current file in the
        Editor plugin.
        """
        plugin = self.focused_plugin
        if plugin and getattr(plugin, 'CAN_HANDLE_FILE_ACTIONS', False):
            plugin.close_file()
        elif self.is_plugin_available(Plugins.Editor):
            editor = self.get_plugin(Plugins.Editor)
            editor.close_file()

    def close_all(self) -> None:
        """
        Close all opened files in the current plugin.

        If the plugin that currently has focus, has its
        `CAN_HANDLE_FILE_ACTIONS` attribute set to `True`, then close all
        files in that plugin. Otherwise, close all files in the Editor plugin.
        """
        plugin = self.focused_plugin
        if plugin and getattr(plugin, 'CAN_HANDLE_FILE_ACTIONS', False):
            plugin.close_all()
        elif self.is_plugin_available(Plugins.Editor):
            editor = self.get_plugin(Plugins.Editor)
            editor.close_all_files()

    def enable_file_action(
        self,
        action_name: str,
        enabled: bool,
        plugin: str
    ) -> None:
        """
        Enable or disable file actions for a given plugin.

        Parameters
        ----------
        action_name : str
            The name of the action to be enabled or disabled. These names
            are listed in ApplicationActions, for instance "New file"
        enabled : bool
            True to enable the action, False to disable it.
        plugin : str
            The name of the plugin for which the file action needs to be
            enabled or disabled.
        """
        self.file_action_enabled[(plugin, action_name)] = enabled
        self._update_file_actions()

    def undo(self) -> None:
        """
        Do an undo operation in the current plugin.

        If the plugin that currently has focus, has its
        `CAN_HANDLE_EDIT_ACTIONS` attribute set to `True`, then do the undo
        operation in that plugin. Otherwise, undo in the Editor plugin.
        """
        plugin = self.focused_plugin
        if plugin and getattr(plugin, 'CAN_HANDLE_EDIT_ACTIONS', False):
            plugin.undo()
        elif self.is_plugin_available(Plugins.Editor):
            editor = self.get_plugin(Plugins.Editor)
            editor.undo()

    def redo(self) -> None:
        """
        Do a redo operation in the current plugin.

        If the plugin that currently has focus, has its
        `CAN_HANDLE_EDIT_ACTIONS` attribute set to `True`, then do the redo
        operation in that plugin. Otherwise, redo in the Editor plugin.
        """
        plugin = self.focused_plugin
        if plugin and getattr(plugin, 'CAN_HANDLE_EDIT_ACTIONS', False):
            plugin.redo()
        elif self.is_plugin_available(Plugins.Editor):
            editor = self.get_plugin(Plugins.Editor)
            editor.redo()

    def cut(self) -> None:
        """
        Do a cut operation in the current plugin.

        If the plugin that currently has focus, has its
        `CAN_HANDLE_EDIT_ACTIONS` attribute set to `True`, then do the cut
        operation in that plugin. Otherwise, cut in the Editor plugin.
        """
        plugin = self.focused_plugin
        if plugin and getattr(plugin, 'CAN_HANDLE_EDIT_ACTIONS', False):
            plugin.cut()
        elif self.is_plugin_available(Plugins.Editor):
            editor = self.get_plugin(Plugins.Editor)
            editor.cut()

    def copy(self) -> None:
        """
        Do a copy operation in the current plugin.

        If the plugin that currently has focus, has its
        `CAN_HANDLE_EDIT_ACTIONS` attribute set to `True`, then do the copy
        operation in that plugin. Otherwise, copy in the Editor plugin.
        """
        plugin = self.focused_plugin
        if plugin and getattr(plugin, 'CAN_HANDLE_EDIT_ACTIONS', False):
            plugin.copy()
        elif self.is_plugin_available(Plugins.Editor):
            editor = self.get_plugin(Plugins.Editor)
            editor.copy()

    def paste(self) -> None:
        """
        Do a paste operation in the current plugin.

        If the plugin that currently has focus, has its
        `CAN_HANDLE_EDIT_ACTIONS` attribute set to `True`, then do the paste
        operation in that plugin. Otherwise, paste in the Editor plugin.
        """
        plugin = self.focused_plugin
        if plugin and getattr(plugin, 'CAN_HANDLE_EDIT_ACTIONS', False):
            plugin.paste()
        elif self.is_plugin_available(Plugins.Editor):
            editor = self.get_plugin(Plugins.Editor)
            editor.paste()

    def select_all(self) -> None:
        """
        Do a select all operation in the current plugin.

        If the plugin that currently has focus, has its
        `CAN_HANDLE_EDIT_ACTIONS` attribute set to `True`, then do the select
        all operation in that plugin. Otherwise, select all in the Editor
        plugin.
        """
        plugin = self.focused_plugin
        if plugin and getattr(plugin, 'CAN_HANDLE_EDIT_ACTIONS', False):
            plugin.select_all()
        elif self.is_plugin_available(Plugins.Editor):
            editor = self.get_plugin(Plugins.Editor)
            editor.select_all()

    def enable_edit_action(
        self,
        action_name: str,
        enabled: bool,
        plugin: str
    ) -> None:
        """
        Enable or disable edit actions for a given plugin.

        Parameters
        ----------
        action_name : str
            The name of the action to be enabled or disabled. These names
            are listed in ApplicationActions, for instance "Undo".
        enabled : bool
            True to enable the action, False to disable it.
        plugin : str
            The name of the plugin for which the edit action needs to be
            enabled or disabled.
        """
        self.edit_action_enabled[(plugin, action_name)] = enabled
        self._update_edit_actions()

    def find(self) -> None:
        """
        Do a find text in the current plugin.

        If the plugin that currently has focus, has its
        `CAN_HANDLE_SEARCH_ACTIONS` attribute set to `True`, then do the find
        text in that plugin. Otherwise, find text in the Editor plugin.
        """
        plugin = self.focused_plugin
        if plugin and getattr(plugin, 'CAN_HANDLE_SEARCH_ACTIONS', False):
            plugin.find()
        elif self.is_plugin_available(Plugins.Editor):
            editor = self.get_plugin(Plugins.Editor)
            editor.find()

    def find_next(self) -> None:
        """
        Do a find next text occurrence in the current plugin.

        If the plugin that currently has focus, has its
        `CAN_HANDLE_SEARCH_ACTIONS` attribute set to `True`, then do the find
        next text occurrence in that plugin. Otherwise, find next text
        occurrence in the Editor plugin.
        """
        plugin = self.focused_plugin
        if plugin and getattr(plugin, 'CAN_HANDLE_SEARCH_ACTIONS', False):
            plugin.find_next()
        elif self.is_plugin_available(Plugins.Editor):
            editor = self.get_plugin(Plugins.Editor)
            editor.find_next()

    def find_previous(self) -> None:
        """
        Do a find previous text occurrence in the current plugin.

        If the plugin that currently has focus, has its
        `CAN_HANDLE_SEARCH_ACTIONS` attribute set to `True`, then do the find
        previous text occurrence in that plugin. Otherwise, find previous text
        occurrence in the Editor plugin.
        """
        plugin = self.focused_plugin
        if plugin and getattr(plugin, 'CAN_HANDLE_SEARCH_ACTIONS', False):
            plugin.find_previous()
        elif self.is_plugin_available(Plugins.Editor):
            editor = self.get_plugin(Plugins.Editor)
            editor.find_previous()

    def replace(self) -> None:
        """
        Do a replace text occurrence in the current plugin.

        If the plugin that currently has focus, has its
        `CAN_HANDLE_SEARCH_ACTIONS` attribute set to `True`, then do the
        replace text occurrence in that plugin. Otherwise, replace text
        occurrence in the Editor plugin.
        """
        plugin = self.focused_plugin
        if plugin and getattr(plugin, 'CAN_HANDLE_SEARCH_ACTIONS', False):
            plugin.replace()
        elif self.is_plugin_available(Plugins.Editor):
            editor = self.get_plugin(Plugins.Editor)
            editor.replace()

    def enable_search_action(
        self,
        action_name: str,
        enabled: bool,
        plugin: str
    ) -> None:
        """
        Enable or disable a search action for a given plugin.

        Parameters
        ----------
        action_name : str
            The name of the action to be enabled or disabled. These names
            are listed in ApplicationActions, for instance "FindText".
        enabled : bool
            True to enable the action, False to disable it.
        plugin : str
            The name of the plugin for which the search action needs to be
            enabled or disabled.
        """
        self.search_action_enabled[(plugin, action_name)] = enabled
        self._update_search_actions()

    @property
    def documentation_action(self):
        """Open Spyder's Documentation in the browser."""
        return self.get_container().documentation_action

    @property
    def video_action(self):
        """Open Spyder's video documentation in the browser."""
        return self.get_container().video_action

    @property
    def trouble_action(self):
        """Open Spyder's troubleshooting documentation in the browser."""
        return self.get_container().trouble_action

    @property
    def dependencies_action(self):
        """Show Spyder's Dependencies dialog box."""
        return self.get_container().dependencies_action

    @property
    def support_group_action(self):
        """Open Spyder's Google support group in the browser."""
        return self.get_container().support_group_action

    @property
    def about_action(self):
        """Show Spyder's About dialog box."""
        return self.get_container().about_action

    @property
    def user_env_action(self):
        """Show Spyder's Windows user env variables dialog box."""
        return self.get_container().user_env_action

    @property
    def restart_action(self):
        """Restart Spyder action."""
        return self.get_container().restart_action

    @property
    def restart_debug_action(self):
        """Restart Spyder in DEBUG mode action."""
        return self.get_container().restart_debug_action

    @property
    def report_action(self):
        """Restart Spyder action."""
        return self.get_container().report_action

    @property
    def debug_logs_menu(self):
        return self.get_container().get_menu(
            ApplicationPluginMenus.DebugLogsMenu)



# -----------------------------------------------------------------------------
# Copyright (c) Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)
# -----------------------------------------------------------------------------

"""
Dialog window for setting the viewing preferences of the data frame editor.
"""

# Standard library import
from typing import Optional

# Third party imports
from qtpy.QtWidgets import (
    QDialog,
    QDialogButtonBox,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QRadioButton,
    QVBoxLayout,
    QWidget
)

# Local imports
from spyder.api.translations import _
from spyder.api.widgets.dialogs import SpyderDialogButtonBox
from spyder.utils.icon_manager import ima
from spyder.widgets.helperwidgets import TipWidget


FORMAT_SPEC_URL = (
    'https://docs.python.org/3/library/string.html'
    '#format-specification-mini-language'
)


class PreferencesDialog(QDialog):
    """
    Dialog window for setting viewing preferences of dataframe or array editor.

    Set the attributes `float_format`, `varying_background` and `global_algo`
    to set the options, if necessary. Call `exec_()` to show the dialog to the
    user and allow them to interact. Finally, read the attributes to retrieve
    the options selected by the user.

    Parameters
    ----------
    type_string: str
        Type of variable being edited; should be 'dataframe' or 'array'.
        The main difference is that arrays do not support the "by column"
        coloring algorithm. Some text also uses the type.
    parent: QWidget, optional
        Parent widget.
    """

    def __init__(self, type_string: str, parent: Optional[QWidget] = None):
        super().__init__(parent)
        self.type_string = type_string

        self.setWindowTitle(
            _('{} editor preferences').format(type_string.capitalize())
        )

        main_layout = QVBoxLayout(self)

        formatting_group = QGroupBox(_('Formatting'))
        formatting_layout = QVBoxLayout(formatting_group)
        format_label = QLabel(
            _('<a href="{url}">Format specification</a> for floats:')
            .format(url=FORMAT_SPEC_URL)
        )

        format_label_layout = QHBoxLayout()
        format_label_layout.setSpacing(0)
        format_label_layout.addWidget(format_label)
        format_tip_text = _(
            'Use same syntax as for built-in <tt>format()</tt> function. '
            'Default is <tt>.6g</tt>.'
        )
        self.add_help_info_label(format_label_layout, format_tip_text)

        # Stylesheet aligns `format_label` and `format_input` to the left
        self.format_input = QLineEdit(self)
        self.format_input.setStyleSheet('margin-left: 5px')

        formatting_layout.addLayout(format_label_layout)
        formatting_layout.addWidget(self.format_input)
        main_layout.addWidget(formatting_group)

        background_group = QGroupBox(_('Background color'))
        background_layout = QVBoxLayout(background_group)

        self.default_background_button = self.create_radio_button(
            _('Use default background color'),
            _('Use same background color for all cells'),
            background_group,
            background_layout
        )

        self.varying_background_button = self.create_radio_button(
            _('Vary background color according to value'),
            _(
                'Use red for largest number, blue for smallest number, '
                'and intermediate colors for the other numbers.'
            ),
            background_group,
            background_layout
        )

        main_layout.addWidget(background_group)

        if type_string == 'dataframe':
            comparator_group = QGroupBox(_('Coloring algorithm'))
            comparator_layout = QVBoxLayout(comparator_group)

            self.global_button = self.create_radio_button(
                _('Global'),
                _(
                    'Compare each cell against the largest and smallest '
                    'numbers in the entire dataframe'
                ),
                comparator_group,
                comparator_layout
            )

            self.by_column_button = self.create_radio_button(
                _('Column by column'),
                _(
                    'Compare each cell against the largest and smallest '
                    'numbers in the same column'
                ),
                comparator_group,
                comparator_layout
            )

            main_layout.addWidget(comparator_group)

        self.buttons = SpyderDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        )
        main_layout.addWidget(self.buttons)
        self.buttons.accepted.connect(self.accept)
        self.buttons.rejected.connect(self.reject)

        self.default_background_button.setChecked(True)

        if type_string == 'dataframe':
            self.global_button.setChecked(True)
            comparator_group.setEnabled(False)
            self.varying_background_button.toggled.connect(
                lambda value: comparator_group.setEnabled(value)
            )

    @property
    def float_format(self) -> str:
        """
        Format specification for floats.
        """
        return self.format_input.text()

    @float_format.setter
    def float_format(self, new_format: str) -> None:
        self.format_input.setText(new_format)

    @property
    def varying_background(self) -> bool:
        """
        Whether to use a colored background.

        If True, then vary the background of the cells in the editor according
        to the value. If False, then use the default background in all cells.
        """
        return self.varying_background_button.isChecked()

    @varying_background.setter
    def varying_background(self, value: bool):
        if value:
            self.varying_background_button.setChecked(True)
        else:
            self.default_background_button.setChecked(True)

    @property
    def global_algo(self) -> bool:
        """
        Whether to use the global minimum and maximum to pick colors.

        If True, then select the background color by comparing the cell value
        against the minimum and maximum over the whole dataframe. If False,
        then use the minimum and maximum over the column.

        This attribute has no effect if `varying_background` is False.
        """
        if self.type_string == 'dataframe':
            return self.global_button.isChecked()
        else:
            return True

    @global_algo.setter
    def global_algo(self, value: bool):
        if self.type_string == 'dataframe':
            if value:
                self.global_button.setChecked(True)
            else:
                self.by_column_button.setChecked(True)

    def add_help_info_label(self, layout, tip_text):
        help_label = TipWidget(
            tip_text=tip_text,
            icon=ima.icon('question_tip'),
            hover_icon=ima.icon('question_tip_hover'),
            wrap_text=True,
        )
        layout.addWidget(help_label)
        layout.addStretch(100)

    def create_radio_button(self, text, tip_text, button_group, layout):
        hor_layout = QHBoxLayout()
        hor_layout.setContentsMargins(0, 0, 0, 0)
        radio_button = QRadioButton(text, button_group)
        hor_layout.addWidget(radio_button)
        self.add_help_info_label(hor_layout, tip_text)
        layout.addLayout(hor_layout)
        return radio_button



# -----------------------------------------------------------------------------
# Copyright (c) 2021- Spyder Project Contributors
#
# Released under the terms of the MIT License
# (see LICENSE.txt in the project root directory for details)
# -----------------------------------------------------------------------------

"""
Enums listing Spyder internal plugins.

.. deprecated:: 6.2

    This module will be renamed to the private :mod:`!_enum` in Spyder 6.2,
    while the current public name will become an alias raising a
    :exc:`DeprecationWarning` on import and be removed in 7.0.

    Use the canonical location, the top-level :mod:`spyder.api.plugins` instead
    which already exports all the public objects of this module.

    For example, ``import spyder.api.plugins`` instead of
    ``import spyder.api.plugins.enums``, or
    ``from spyder.api.plugins import Plugins`` instead of
    ``from spyder.api.plugins.enums import Plugins``.
"""


class DockablePlugins:
    """Pseudo-enum class listing the names of Spyder internal dockable plugins.

    Values correspond to the plugin's
    :attr:`~spyder.api.plugins.SpyderPluginV2.NAME`.
    """

    Console = "internal_console"
    """The Spyder Internal Console plugin."""

    Debugger = "debugger"
    """The Spyder Debugger plugin."""

    Editor = "editor"
    """The Spyder Editor plugin."""

    Explorer = "explorer"
    """The Spyder Files plugin."""

    Find = "find_in_files"
    """The Spyder Find plugin."""

    Help = "help"
    """The Spyder Help plugin."""

    History = "historylog"
    """The Spyder History plugin."""

    IPythonConsole = "ipython_console"
    """The Spyder IPython Console plugin."""

    OnlineHelp = "onlinehelp"
    """The Spyder Online Help plugin."""

    OutlineExplorer = "outline_explorer"
    """The Spyder Outline plugin."""

    Plots = "plots"
    """The Spyder Plots plugin."""

    Profiler = "profiler"
    """The Spyder Profiler plugin."""

    Projects = "project_explorer"
    """The Spyder Projects plugin."""

    Pylint = "pylint"
    """The Spyder Code Analysis plugin."""

    VariableExplorer = "variable_explorer"
    """The Spyder Variable Explorer plugin."""


class Plugins(DockablePlugins):
    """Pseudo-enum class listing the names of all Spyder internal plugins.

    Values correspond to the plugin's
    :attr:`~spyder.api.plugins.SpyderPluginV2.NAME`.

    .. note::

        This "enum" also includes all the members from :class:`DockablePlugins`
        via inheritance from that class.
    """

    All = "all"
    """Wildcard to populate :attr:`~spyder.api.plugins.SpyderPluginV2.REQUIRES`
    with all available plugins.
    """

    Appearance = "appearance"
    """The Spyder Appearance plugin."""

    Application = "application"
    """The Spyder Application plugin."""

    Completions = "completions"
    """The Spyder Completions plugin."""

    ExternalTerminal = "external_terminal"
    """The Spyder External Terminal plugin."""

    Layout = "layout"
    """The Spyder Layout plugin."""

    MainInterpreter = "main_interpreter"
    """The Spyder Main Interpreter plugin."""

    MainMenu = "mainmenu"
    """The Spyder Main Menu plugin."""

    Preferences = "preferences"
    """The Spyder Preferences plugin."""

    PythonpathManager = "pythonpath_manager"
    """The Spyder PYTHONPATH Manager plugin."""

    RemoteClient = "remoteclient"
    """The Spyder Remote Client plugin."""

    Run = "run"
    """The Spyder Run plugin."""

    Shortcuts = "shortcuts"
    """The Spyder Shortcuts plugin."""

    StatusBar = "statusbar"
    """The Spyder Status Bar plugin."""

    Switcher = "switcher"
    """The Spyder Switcher plugin."""

    Toolbar = "toolbar"
    """The Spyder Toolbar plugin."""

    Tours = "tours"
    """The Spyder Tours plugin."""

    UpdateManager = "update_manager"
    """The Spyder Update Manager plugin."""

    WorkingDirectory = "workingdir"
    """The Spyder Working Directory plugin."""


class OptionalPlugins:
    """Pseudo-enum class listing the names of Spyder optional plugins.

    Values correspond to the plugin's
    :attr:`~spyder.api.plugins.SpyderPluginV2.NAME`.
    """

    EnvManager = "spyder_env_manager"
    """The Spyder Environment Manager plugin."""



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Language Server Protocol configuration tabs."""

from .advanced import AdvancedConfigTab
from .formatting import FormattingConfigTab
from .introspection import IntrospectionConfigTab
from .linting import LintingConfigTab
from .otherlanguages import OtherLanguagesConfigTab


# LSP provider tabs
TABS = [
    LintingConfigTab,
    IntrospectionConfigTab,
    FormattingConfigTab,
    AdvancedConfigTab,
    OtherLanguagesConfigTab
]



# -----------------------------------------------------------------------------
# Copyright (c) 2021- Spyder Project Contributors
#
# Released under the terms of the MIT License
# (see LICENSE.txt in the project root directory for details)
# -----------------------------------------------------------------------------

"""
Classes to connect a plugin to the shell widgets in the IPython console.
"""



# -----------------------------------------------------------------------------
# Copyright (c) 2020- Spyder Project Contributors
#
# Released under the terms of the MIT License
# (see LICENSE.txt in the project root directory for details)
# -----------------------------------------------------------------------------

"""
Spyder API menu widgets.
"""

from __future__ import annotations

# Standard library imports
import sys
from typing import TypeVar, TYPE_CHECKING

# Third party imports
import qstylizer.style
from qtpy.QtCore import QTimer
from qtpy.QtGui import QCursor
from qtpy.QtWidgets import QAction, QMenu, QProxyStyle, QStyle

# Local imports
from spyder.api.fonts import SpyderFontType, SpyderFontsMixin
from spyder.utils.qthelpers import add_actions, set_menu_icons, SpyderAction
from spyder.utils.palette import SpyderPalette
from spyder.utils.stylesheet import AppStyle, MAC, WIN

if TYPE_CHECKING:
    from qtpy.QtGui import QShowEvent
    from qtpy.QtWidgets import QStyleOption, QWidget

    import spyder.utils.qthelpers  # For SpyderAction


# ---- Constants
# -----------------------------------------------------------------------------
MENU_SEPARATOR: None = None
"""Constant representing a separator line between groups in a menu."""


# Generic type annotations
T = TypeVar("T", bound="SpyderMenu")
"""Generic :class:`SpyderMenu` type variable (:class:`~typing.TypeVar`)."""


class OptionsMenuSections:
    """Pseudo-enum listing sections in the pane options (hamburger) menu."""

    Top: str = "top_section"
    """The top section of the options menu."""

    Bottom: str = "bottom_section"
    """The bottom section of the options menu."""


class PluginMainWidgetMenus:
    """Pseudo-enum listing the menu types dockable plugin main widgets have."""

    Context: str = "context_menu"
    """The widget context menu."""

    Options: str = "options_menu"
    """The widget options (hamburger) menu."""


# ---- Style
# -----------------------------------------------------------------------------
class SpyderMenuProxyStyle(QProxyStyle):
    """Menu style adjustments that can only be done with a proxy style.

    .. deprecated:: 6.2

        This class will be moved to the private :class:`!_SpyderMenuProxyStyle`
        in Spyder 6.2, while the current public name will become an alias
        raising a :exc:`DeprecationWarning` on use, and removed in 7.0.

        It was never intended to be used directly by plugins, and its
        functionality is automatically inherited by using :class:`SpyderMenu`.
    """

    def pixelMetric(
        self,
        metric: QStyle.PixelMetric,
        option: QStyleOption | None = None,
        widget: QWidget | None = None,
    ) -> int:
        """
        Calculate the value of the given pixel metric.

        This is a callback intended to be called internally by Qt.

        Parameters
        ----------
        metric : QStyle.PixelMetric
            The pixel metric to calculate.
        option : QStyleOption | None, optional
            The current style options, or ``None`` (default).
        widget : QWidget | None, optional
            The widget the pixel metric will be used for,
            or ``None`` (default).

        Returns
        -------
        int
            The resulting pixel metric value, used internally by Qt.
        """
        if metric == QStyle.PM_SmallIconSize:
            # Change icon size for menus.
            # Taken from https://stackoverflow.com/a/42145885/438386
            delta = -1 if MAC else (0 if WIN else 1)

            return (
                QProxyStyle.pixelMetric(self, metric, option, widget) + delta
            )

        return QProxyStyle.pixelMetric(self, metric, option, widget)


# ---- Widgets
# -----------------------------------------------------------------------------
class SpyderMenu(QMenu, SpyderFontsMixin):
    """
    A QMenu subclass implementing additional functionality for Spyder.

    All menus in Spyder must inherit from this class.
    """

    MENUS: list[tuple[QWidget | None, str | None, SpyderMenu]] = []
    """
    List of 3-tuples, one per menu, describing each menu.

    Contains the menu's parent widget object, its name as a string,
    and the :class:`SpyderMenu` object itself, in that order.
    """

    APP_MENU: bool = False
    """
    Whether this is a main application menu or a plugin menu.

    Set this to ``True`` if this is an application menu; ``False`` otherwise.
    """

    HORIZONTAL_MARGIN_FOR_ITEMS: int = 2 * AppStyle.MarginSize
    """The QSS horizontal (left/right) margin for menu items, in pixels."""

    HORIZONTAL_PADDING_FOR_ITEMS: int = 3 * AppStyle.MarginSize
    """The QSS horizontal (left/right) padding for menu items, in pixels."""

    def __init__(
        self,
        parent: QWidget | None = None,
        menu_id: str | None = None,
        title: str | None = None,
        min_width: int | None = None,
        reposition: bool = True,
    ):
        """
        Create a menu for Spyder.

        Parameters
        ----------
        parent: QWidget | None, optional
            The menu's parent widget, or ``None`` if no parent.
        menu_id: str | None, optional
            Unique string identifier for the menu, or ``None`` if no ID.
        title: str | None, optional
            Localized title for the menu, or ``None`` (default) if no title.
        min_width: int or None, optional
            Minimum width for the menu, or ``None`` (default) for no min width.
        reposition: bool, optional
            Whether to vertically reposition the menu due to its padding.
            ``True`` by default.

        Returns
        -------
        None
        """
        self.menu_id: str | None = menu_id
        """The unique string identifier for the menu (or ``None`` if unset)."""

        self._parent = parent
        self._title = title
        self._reposition = reposition

        self._sections = []
        self._actions = []
        self._actions_map = {}
        self._unintroduced_actions = {}
        self._after_sections = {}
        self._dirty = False
        self._is_shown = False
        self._is_submenu = False
        self._in_app_menu = False

        if title is None:
            super().__init__(parent)
        else:
            super().__init__(title, parent)

        self.MENUS.append((parent, title, self))

        # Set min width
        if min_width is not None:
            self.setMinimumWidth(min_width)

        # Signals
        self.aboutToShow.connect(self.render)

        # Adjustments for Mac
        if sys.platform == "darwin":
            # Needed to enable the dynamic population of actions in app menus
            # in the aboutToShow signal.
            # See spyder-ide/spyder#14612
            if self.APP_MENU:
                self.addAction(QAction(self))

            # Necessary to follow Mac's HIG for app menus.
            self.aboutToShow.connect(self._set_icons)

        # Style
        self.css = self._generate_stylesheet()
        self.setStyleSheet(self.css.toString())

        style = SpyderMenuProxyStyle(None)
        style.setParent(self)
        self.setStyle(style)

    # ---- Public API
    # -------------------------------------------------------------------------
    def clear_actions(self) -> None:
        """
        Remove actions from the menu (including custom references).

        Returns
        -------
        None
        """
        self.clear()
        self._sections = []
        self._actions = []
        self._actions_map = {}
        self._unintroduced_actions = {}
        self._after_sections = {}

    def add_action(
        self,
        action: spyder.utils.qthelpers.SpyderAction | SpyderMenu,
        section: str | None = None,
        before: str | None = None,
        before_section: str | None = None,
        check_before: bool = True,
        omit_id: bool = False,
    ) -> None:
        """
        Add action to a given menu section.

        Parameters
        ----------
        action: spyder.utils.qthelpers.SpyderAction | SpyderMenu
            The action or menu object to add to the menu.
        section: str | None, optional
            The section id in which to insert the ``action``, or ``None``
            (default) for no section.
        before: str | None, optional
            Make the ``action`` appear before the action with the identifier
            ``before``. If ``None`` (default), add it to the end.
        before_section: str | None, optional
            Make the item section appear prior to ``before_section``.
            If ``None`` (the default), add the section to the end.
        check_before: bool, optional
            Check if the ``before`` action is already part of the menu
            before adding this one, and if so save it to be added later.
            This is necessary to avoid an infinite recursion when adding
            unintroduced actions with this method again. ``True`` by default.
        omit_id: bool, optional
            If ``False``, the default, then the menu will check if
            ``action.action_id`` exists and is set to a string, and raise
            an :exc:`AttributeError` if either is not the case.
            If ``True``, it will add the ``action`` anyway.

        Returns
        -------
        None

        Raises
        ------
        AttributeError
            If ``omit_id`` is ``False`` (the default) and ``action.action_id``
            does not exist or is not set to a string.
        """
        item_id = None
        if isinstance(action, SpyderAction) or hasattr(action, "action_id"):
            item_id = action.action_id

            # This is necessary when we set a menu for `action`, e.g. for
            # todo_list_action in EditorMainWidget.
            if action.menu() and isinstance(action.menu(), SpyderMenu):
                action.menu()._is_submenu = True
        elif isinstance(action, SpyderMenu) or hasattr(action, "menu_id"):
            item_id = action.menu_id
            action._is_submenu = True

        if not omit_id and item_id is None and action is not None:
            raise AttributeError(f"Item {action} must declare an id.")

        if before is None:
            self._actions.append((section, action))
        else:
            new_actions = []
            added = False
            before_item = self._actions_map.get(before, None)

            for sec, act in self._actions:
                if before_item is not None and act == before_item:
                    added = True
                    new_actions.append((section, action))

                new_actions.append((sec, act))

            # Actions can't be added to the menu if the `before` action is
            # not part of it yet. That's why we need to save them in the
            # `_unintroduced_actions` dict, so we can add them again when
            # the menu is rendered.
            if not added and check_before:
                before_actions = self._unintroduced_actions.get(before, [])
                before_actions.append((section, action))
                self._unintroduced_actions[before] = before_actions

            self._actions = new_actions

        if section not in self._sections:
            self._add_section(section, before_section)

        # Track state of menu to avoid re-rendering if menu has not changed
        self._dirty = True
        self._actions_map[item_id] = action

    def remove_action(self, item_id: str) -> None:
        """
        Remove the action with the given string identifier.

        Parameters
        ----------
        item_id : str
            The string identifier of the action to remove.

        Returns
        -------
        None
        """
        if item_id in self._actions_map:
            action = self._actions_map.pop(item_id)
            position = None

            for i, (_, act) in enumerate(self._actions):
                if act == action:
                    position = i
                    break

            if position is not None:
                self._actions.pop(position)
                self._dirty = True

    def get_title(self) -> str | None:
        """
        Return the title for the menu.

        Returns
        -------
        str | None
            The menu's title, or ``None`` if it doesn't have one set.
        """
        return self._title

    def get_actions(
        self,
    ) -> list[SpyderMenu | spyder.utils.qthelpers.SpyderAction]:
        """
        Return a parsed list of menu items/actions.

        Includes a :const:`MENU_SEPARATOR` between each defined menu section.

        Returns
        -------
        list[SpyderMenu | spyder.utils.qthelpers.SpyderAction]
            The list of menu items/actions for this menu.
        """
        actions = []
        for section in self._sections:
            for sec, action in self._actions:
                if sec == section:
                    actions.append(action)

            actions.append(MENU_SEPARATOR)
        return actions

    def get_sections(self) -> tuple[str, ...]:
        """
        Return a tuple of menu section names.

        Returns
        -------
        tuple[str, ...]
            An arbitrary-length tuple of menu section names.
        """
        return tuple(self._sections)

    # ---- Private API
    # -------------------------------------------------------------------------
    def _add_missing_actions(self) -> None:
        """
        Add actions that were not introduced to the menu because a `before`
        action they require is not part of it.
        """
        for before, actions in self._unintroduced_actions.items():
            for section, action in actions:
                self.add_action(
                    action, section=section, before=before, check_before=False
                )

        self._unintroduced_actions = {}

    def render(self, force: bool = False) -> None:
        """
        Create the menu prior to showing it. This takes into account sections
        and location of menus.

        Parameters
        ----------
        force: bool, optional
            Whether to force rendering the menu.
        """
        if self._dirty or force:
            self.clear()
            self._add_missing_actions()

            actions = self.get_actions()
            add_actions(self, actions)
            self._set_icons()

            self._dirty = False

    def _add_section(
        self, section: str, before_section: str | None = None
    ) -> None:
        """
        Add a new section to the list of sections in this menu.

        Parameters
        ----------
        section: str
            The name of the section to add.
        before_section: str | None, optional
            Make `section` appear before another one.
        """
        inserted_before_other = False

        if before_section is not None:
            if before_section in self._sections:
                # If before_section was already introduced, we simply need to
                # insert the new section on its position, which will put it
                # exactly behind before_section.
                idx = self._sections.index(before_section)
                self._sections.insert(idx, section)
                inserted_before_other = True
            else:
                # If before_section hasn't been introduced yet, we know we need
                # to insert it after section when it's finally added to the
                # menu. So, we preserve that info in the _after_sections dict.
                self._after_sections[before_section] = section

                # Append section to the list of sections because we assume
                # people build menus from top to bottom, i.e. they add its
                # upper sections first.
                self._sections.append(section)
        else:
            self._sections.append(section)

        # Check if section should be inserted after another one, according to
        # what we have in _after_sections.
        after_section = self._after_sections.pop(section, None)

        if after_section is not None:
            if not inserted_before_other:
                # Insert section to the right of after_section, if it was not
                # inserted before another one.
                if section in self._sections:
                    self._sections.remove(section)

                index = self._sections.index(after_section)
                self._sections.insert(index + 1, section)
            else:
                # If section was already inserted before another one, then we
                # need to move after_section to its left instead.
                if after_section in self._sections:
                    self._sections.remove(after_section)

                idx = self._sections.index(section)
                self._sections.insert(idx, after_section)

    def _set_icons(self) -> None:
        """
        Unset menu icons for app menus and set them for regular menus.

        This is necessary only for Mac to follow its Human Interface
        Guidelines (HIG), which don't recommend icons in app menus.
        """
        if sys.platform == "darwin":
            if self.APP_MENU or self._in_app_menu:
                set_menu_icons(self, False, in_app_menu=True)
            else:
                set_menu_icons(self, True)

    @classmethod
    def _generate_stylesheet(cls) -> qstylizer.style.StyleSheet:
        """Generate base stylesheet for menus."""
        css = qstylizer.style.StyleSheet()
        font = cls.get_font(SpyderFontType.Interface)

        # Add padding and border to follow modern standards
        css.QMenu.setValues(
            # Only add top and bottom padding so that menu separators can go
            # completely from the left to right border.
            paddingTop=f"{2 * AppStyle.MarginSize}px",
            paddingBottom=f"{2 * AppStyle.MarginSize}px",
            # This uses the same color as the separator
            border=f"1px solid {SpyderPalette.COLOR_BACKGROUND_6}",
        )

        # Set the right background color. This is the only way to do it!
        css["QWidget:disabled QMenu"].setValues(
            backgroundColor=SpyderPalette.COLOR_BACKGROUND_3,
        )

        # Add padding around separators to prevent that hovering on items hides
        # them.
        css["QMenu::separator"].setValues(
            # Only add top and bottom margins so that the separators can go
            # completely from the left to right border.
            margin=f"{2 * AppStyle.MarginSize}px 0px",
        )

        # Set menu item properties
        delta_top = 0 if (MAC or WIN) else 1
        delta_bottom = 0 if MAC else (2 if WIN else 1)
        css["QMenu::item"].setValues(
            height="1.1em" if MAC else ("1.35em" if WIN else "1.25em"),
            marginLeft=f"{cls.HORIZONTAL_MARGIN_FOR_ITEMS}px",
            marginRight=f"{cls.HORIZONTAL_MARGIN_FOR_ITEMS}px",
            paddingTop=f"{AppStyle.MarginSize + delta_top}px",
            paddingBottom=f"{AppStyle.MarginSize + delta_bottom}px",
            paddingLeft=f"{cls.HORIZONTAL_PADDING_FOR_ITEMS}px",
            paddingRight=f"{cls.HORIZONTAL_PADDING_FOR_ITEMS}px",
            fontFamily=font.family(),
            fontSize=f"{font.pointSize()}pt",
            backgroundColor="transparent",
        )

        # Set hover and pressed state of items
        for state in ["selected", "pressed"]:
            if state == "selected":
                bg_color = SpyderPalette.COLOR_BACKGROUND_4
            else:
                bg_color = SpyderPalette.COLOR_BACKGROUND_5

            css[f"QMenu::item:{state}"].setValues(
                backgroundColor=bg_color,
                borderRadius=SpyderPalette.SIZE_BORDER_RADIUS,
            )

        # Set disabled state of items
        for state in ["disabled", "selected:disabled"]:
            css[f"QMenu::item:{state}"].setValues(
                color=SpyderPalette.COLOR_DISABLED,
                backgroundColor="transparent",
            )

        return css

    def __str__(self) -> str:
        """
        Output this menu's class name and identifier as a string.

        Returns
        -------
        str
            The menu's class name and string identifier, in the format
            :file:``SpyderMenu({MENU_ID})``.
        """
        return f"SpyderMenu('{self.menu_id}')"

    def __repr__(self) -> str:
        """
        Output this menu's class name and identifier as a string.

        Returns
        -------
        str
            The menu's class name and string identifier, in the format
            :file:``SpyderMenu({MENU_ID})``.
        """
        return f"SpyderMenu('{self.menu_id}')"

    def _adjust_menu_position(self) -> None:
        """Menu position adjustment logic to follow custom style."""
        if not self._is_shown:
            # Reposition submenus vertically due to padding and border
            if self._reposition and self._is_submenu:
                self.move(
                    self.pos().x(),
                    # Current vertical pos - padding - border
                    self.pos().y() - 2 * AppStyle.MarginSize - 1,
                )

            self._is_shown = True

        # Reposition menus horizontally due to border
        if self.APP_MENU:
            delta_x = 0 if MAC else 3
        else:
            if QCursor().pos().x() - self.pos().x() < 40:
                # If the difference between the current cursor x position and
                # the menu one is small, it means the menu will be shown to the
                # right, so we need to move it in that direction.
                delta_x = 1
            else:
                # This happens when the menu is shown to the left.
                delta_x = -1

        self.move(self.pos().x() + delta_x, self.pos().y())

    # ---- Qt methods
    # -------------------------------------------------------------------------
    def showEvent(self, event: QShowEvent) -> None:
        """
        Perform adjustments when the menu is going to be shown.

        Parameters
        ----------
        event : QShowEvent
            The event object showing the menu.

        Returns
        -------
        None
        """
        # To prevent race conditions which can cause partially showing a menu
        # (as in spyder-ide/spyder#22266), we use a timer to queue the move
        # related events after the menu is shown.
        # For more info you can check:
        #  * https://forum.qt.io/topic/23381/showevent-not-working/3
        #  * https://stackoverflow.com/a/49351518
        QTimer.singleShot(0, self._adjust_menu_position)


class PluginMainWidgetOptionsMenu(SpyderMenu):
    """
    Options menu for :class:`~spyder.api.widgets.main_widget.PluginMainWidget`.
    """

    def render(self) -> None:
        """
        Render the menu's bottom section as expected.

        Returns
        -------
        None
        """
        if self._dirty:
            self.clear()
            self._add_missing_actions()

            bottom = OptionsMenuSections.Bottom
            actions = []
            for section in self._sections:
                for sec, action in self._actions:
                    if sec == section and sec != bottom:
                        actions.append(action)

                actions.append(MENU_SEPARATOR)

            # Add bottom actions
            for sec, action in self._actions:
                if sec == bottom:
                    actions.append(action)

            add_actions(self, actions)
            self._set_icons()

            self._dirty = False
